"""Strategies for customizing subclass behaviors."""
from gc import collect
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union

from ..converters import BaseConverter, Converter
from ..gen import AttributeOverride, make_dict_structure_fn, make_dict_unstructure_fn
from ..gen._consts import already_generating


def _make_subclasses_tree(cl: Type) -> List[Type]:
    return [cl] + [
        sscl for scl in cl.__subclasses__() for sscl in _make_subclasses_tree(scl)
    ]


def _has_subclasses(cl: Type, given_subclasses: Tuple[Type, ...]) -> bool:
    """Whether the given class has subclasses from `given_subclasses`."""
    actual = set(cl.__subclasses__())
    given = set(given_subclasses)
    return bool(actual & given)


def _get_union_type(cl: Type, given_subclasses_tree: Tuple[Type]) -> Optional[Type]:
    actual_subclass_tree = tuple(_make_subclasses_tree(cl))
    class_tree = tuple(set(actual_subclass_tree) & set(given_subclasses_tree))
    if len(class_tree) >= 2:
        union_type = Union[class_tree]
    else:
        union_type = None
    return union_type


def include_subclasses(
    cl: Type,
    converter: Converter,
    subclasses: Optional[Tuple[Type, ...]] = None,
    union_strategy: Optional[Callable[[Any, BaseConverter], Any]] = None,
    overrides: Optional[Dict[str, AttributeOverride]] = None,
) -> None:
    """
    Configure the converter so that the attrs/dataclass `cl` is un/structured as if it
    was a union of itself and all its subclasses that are defined at the time when this
    strategy is applied.

    :param cl: A base `attrs` or `dataclass` class.
    :param converter: The `Converter` on which this strategy is applied. Do note that
        the strategy does not work for a :class:`cattrs.BaseConverter`.
    :param subclasses: A tuple of sublcasses whose ancestor is `cl`. If left as `None`,
        subclasses are detected using recursively the `__subclasses__` method of `cl`
        and its descendents.
    :param union_strategy: A callable of two arguments passed by position
        (`subclass_union`, `converter`) that defines the union strategy to use to
        disambiguate the subclasses union. If `None` (the default), the automatic unique
        field disambiguation is used which means that every single subclass
        participating in the union must have an attribute name that does not exist in
        any other sibling class.
    :param overrides: a mapping of `cl` attribute names to overrides (instantiated with
        :func:`cattrs.gen.override`) to customize un/structuring.

    .. versionadded:: 23.1.0
    """
    # Due to https://github.com/python-attrs/attrs/issues/1047
    collect()
    if subclasses is not None:
        parent_subclass_tree = (cl,) + subclasses
    else:
        parent_subclass_tree = tuple(_make_subclasses_tree(cl))

    if overrides is None:
        overrides = {}

    if union_strategy is None:
        _include_subclasses_without_union_strategy(
            cl, converter, parent_subclass_tree, overrides
        )
    else:
        _include_subclasses_with_union_strategy(
            converter, parent_subclass_tree, union_strategy, overrides
        )


def _include_subclasses_without_union_strategy(
    cl,
    converter: Converter,
    parent_subclass_tree: Tuple[Type],
    overrides: Dict[str, AttributeOverride],
):
    # The iteration approach is required if subclasses are more than one level deep:
    for cl in parent_subclass_tree:
        # We re-create a reduced union type to handle the following case:
        #
        #     converter.structure(d, as=Child)
        #
        # In the above, the `as=Child` argument will be transformed to a union type of
        # itself and its subtypes, that way we guarantee that the returned object will
        # not be the parent.
        subclass_union = _get_union_type(cl, parent_subclass_tree)

        def cls_is_cl(cls, _cl=cl):
            return cls is _cl

        base_struct_hook = make_dict_structure_fn(cl, converter, **overrides)
        base_unstruct_hook = make_dict_unstructure_fn(cl, converter, **overrides)

        if subclass_union is None:

            def struct_hook(val: dict, _, _cl=cl, _base_hook=base_struct_hook) -> cl:
                return _base_hook(val, _cl)

        else:
            dis_fn = converter._get_dis_func(subclass_union)

            def struct_hook(
                val: dict,
                _,
                _c=converter,
                _cl=cl,
                _base_hook=base_struct_hook,
                _dis_fn=dis_fn,
            ) -> cl:
                """
                If val is disambiguated to the class `cl`, use its base hook.

                If val is disambiguated to a subclass, dispatch on its exact runtime
                type.
                """
                dis_cl = _dis_fn(val)
                if dis_cl is _cl:
                    return _base_hook(val, _cl)
                return _c.structure(val, dis_cl)

        def unstruct_hook(
            val: parent_subclass_tree[0],
            _c=converter,
            _cl=cl,
            _base_hook=base_unstruct_hook,
        ) -> Dict:
            """
            If val is an instance of the class `cl`, use the hook.

            If val is an instance of a subclass, dispatch on its exact runtime type.
            """
            if val.__class__ is _cl:
                return _base_hook(val)
            return _c.unstructure(val, unstructure_as=val.__class__)

        # This needs to use function dispatch, using singledispatch will again
        # match A and all subclasses, which is not what we want.
        converter.register_structure_hook_func(cls_is_cl, struct_hook)
        converter.register_unstructure_hook_func(cls_is_cl, unstruct_hook)


def _include_subclasses_with_union_strategy(
    converter: Converter,
    union_classes: Tuple[Type, ...],
    union_strategy: Callable[[Any, BaseConverter], Any],
    overrides: Dict[str, AttributeOverride],
):
    """
    This function is tricky because we're dealing with what is essentially a circular reference.

    We need to generate a structure hook for a class that is both:
    * specific for that particular class and its own fields
    * but should handle specific functions for all its descendants too

    Hence the dance with registering below.
    """

    parent_classes = [cl for cl in union_classes if _has_subclasses(cl, union_classes)]
    if not parent_classes:
        return

    original_unstruct_hooks = {}
    original_struct_hooks = {}
    for cl in union_classes:
        # In the first pass, every class gets its own unstructure function according to
        # the overrides.
        # We just generate the hooks, and do not register them. This allows us to manipulate
        # the _already_generating set to force runtime dispatch.
        already_generating.working_set = set(union_classes) - {cl}
        try:
            unstruct_hook = make_dict_unstructure_fn(cl, converter, **overrides)
            struct_hook = make_dict_structure_fn(cl, converter, **overrides)
        finally:
            already_generating.working_set = set()
        original_unstruct_hooks[cl] = unstruct_hook
        original_struct_hooks[cl] = struct_hook

    # Now that's done, we can register all the hooks and generate the
    # union handler. The union handler needs them.
    final_union = Union[union_classes]  # type: ignore

    for cl, hook in original_unstruct_hooks.items():

        def cls_is_cl(cls, _cl=cl):
            return cls is _cl

        converter.register_unstructure_hook_func(cls_is_cl, hook)

    for cl, hook in original_struct_hooks.items():

        def cls_is_cl(cls, _cl=cl):
            return cls is _cl

        converter.register_structure_hook_func(cls_is_cl, hook)

    union_strategy(final_union, converter)
    unstruct_hook = converter._unstructure_func.dispatch(final_union)
    struct_hook = converter._structure_func.dispatch(final_union)

    for cl in union_classes:
        # In the second pass, we overwrite the hooks with the union hook.

        def cls_is_cl(cls, _cl=cl):
            return cls is _cl

        converter.register_unstructure_hook_func(cls_is_cl, unstruct_hook)
        subclasses = tuple([c for c in union_classes if issubclass(c, cl)])
        if len(subclasses) > 1:
            u = Union[subclasses]  # type: ignore
            union_strategy(u, converter)
            struct_hook = converter._structure_func.dispatch(u)

            def sh(payload: dict, _, _u=u, _s=struct_hook) -> cl:
                return _s(payload, _u)

            converter.register_structure_hook_func(cls_is_cl, sh)



"""`gen` tests under PEP 563."""
from __future__ import annotations

from dataclasses import dataclass

from attr import define

from cattrs import Converter
from cattrs.gen import make_dict_structure_fn, make_dict_unstructure_fn


# These need to be at the top level for `attr.resolve_types` to work.
@define
class Inner:
    a: int


@define
class Outer:
    inner: Inner


@define
class InnerB:
    a: int


@define
class OuterB:
    inner: InnerB


@dataclass
class InnerDataclass:
    a: int


@dataclass
class OuterDataclass:
    inner: InnerDataclass


def test_roundtrip():
    converter = Converter()

    fn = make_dict_unstructure_fn(Outer, converter)

    inst = Outer(Inner(1))

    converter.register_unstructure_hook(Outer, fn)

    res_actual = converter.unstructure(inst)

    assert {"inner": {"a": 1}} == res_actual

    converter.register_structure_hook(OuterB, make_dict_structure_fn(OuterB, converter))

    assert converter.structure({"inner": {"a": 1}}, OuterB) == OuterB(InnerB(1))


def test_roundtrip_dc():
    converter = Converter()

    fn = make_dict_unstructure_fn(OuterDataclass, converter)
    converter.register_unstructure_hook(OuterDataclass, fn)

    inst = OuterDataclass(InnerDataclass(1))

    res_actual = converter.unstructure(inst)

    assert {"inner": {"a": 1}} == res_actual

    converter.register_structure_hook(
        OuterDataclass, make_dict_structure_fn(OuterDataclass, converter)
    )

    assert converter.structure({"inner": {"a": 1}}, OuterDataclass) == OuterDataclass(
        InnerDataclass(1)
    )



"""Tests for the factory hooks documentation."""
from attr import define, fields, has

from cattrs.gen import make_dict_structure_fn, make_dict_unstructure_fn, override


def to_camel_case(snake_str):
    components = snake_str.split("_")
    return components[0] + "".join(x.title() for x in components[1:])


def test_snake_to_camel(converter_cls):
    @define
    class Inner:
        a_snake_case_int: int
        a_snake_case_float: float
        a_snake_case_str: str

    @define
    class Outer:
        a_snake_case_inner: Inner

    converter = converter_cls()

    def unstructure_adapt_to_camel_case(type):
        return make_dict_unstructure_fn(
            type,
            converter,
            **{a.name: override(rename=to_camel_case(a.name)) for a in fields(type)}
        )

    converter.register_unstructure_hook_factory(has, unstructure_adapt_to_camel_case)

    original = Outer(Inner(0, 0.0, "str"))
    unstructured = converter.unstructure(original)

    assert unstructured == {
        "aSnakeCaseInner": {
            "aSnakeCaseInt": 0,
            "aSnakeCaseFloat": 0.0,
            "aSnakeCaseStr": "str",
        }
    }

    def structure_adapt_to_camel_case(type):
        overrides = {
            a.name: override(rename=to_camel_case(a.name)) for a in fields(type)
        }
        return make_dict_structure_fn(type, converter, **overrides)

    converter.register_structure_hook_factory(has, structure_adapt_to_camel_case)

    structured = converter.structure(unstructured, Outer)
    assert structured == original



from cattrs.dispatch import FunctionDispatch


def test_function_dispatch():
    dispatch = FunctionDispatch()

    assert dispatch.dispatch(float) is None

    test_func = object()

    dispatch.register(lambda cls: issubclass(cls, float), test_func)

    assert dispatch.dispatch(float) == test_func


def test_function_clears_cache_after_function_added():
    dispatch = FunctionDispatch()

    class Foo(object):
        pass

    Foo()

    class Bar(Foo):
        pass

    Bar()

    dispatch.register(lambda cls: issubclass(cls, Foo), "foo")
    assert dispatch.dispatch(Bar) == "foo"
    dispatch.register(lambda cls: issubclass(cls, Bar), "bar")
    assert dispatch.dispatch(Bar) == "bar"



import collections
import typing

import attr
import pytest

from cattrs import BaseConverter


def test_inheritance(converter):
    @attr.define
    class A:
        i: int

    @attr.define
    class B(A):
        j: int

    assert A(1) == converter.structure(dict(i=1), A)
    assert B(1, 2) == converter.structure(dict(i=1, j=2), B)


def test_gen_hook_priority(converter: BaseConverter):
    """Autogenerated hooks should not take priority over manual hooks."""

    @attr.define
    class A:
        i: int

    @attr.define
    class B(A):
        pass

    # This will generate a hook.
    assert converter.structure({"i": 1}, B) == B(1)

    # Now we register a manual hook for the superclass.
    converter.register_structure_hook(A, lambda o, T: T(o["i"] + 1))

    # This should still work, but using the new hook instead.
    assert converter.structure({"i": 1}, B) == B(2)


@pytest.mark.parametrize(
    "typing_cls", [typing.Hashable, typing.Iterable, typing.Reversible]
)
def test_inherit_typing(converter: BaseConverter, typing_cls):
    """Stuff from typing.* resolves to runtime to collections.abc.*.

    Hence, typing.* are of a special alias type which we want to check if
    cattrs handles them correctly.
    """

    @attr.define
    class A(typing_cls):
        i: int = 0

        def __hash__(self):
            return hash(self.i)

        def __iter__(self):
            return iter([self.i])

        def __reversed__(self):
            return iter([self.i])

    assert converter.structure({"i": 1}, A) == A(i=1)


@pytest.mark.parametrize(
    "collections_abc_cls",
    [collections.abc.Hashable, collections.abc.Iterable, collections.abc.Reversible],
)
def test_inherit_collections_abc(converter: BaseConverter, collections_abc_cls):
    """As extension of test_inherit_typing, check if collections.abc.* work."""

    @attr.define
    class A(collections_abc_cls):
        i: int = 0

        def __hash__(self):
            return hash(self.i)

        def __iter__(self):
            return iter([self.i])

        def __reversed__(self):
            return iter([self.i])

    assert converter.structure({"i": 1}, A) == A(i=1)



"""Preconfigured converters for the stdlib json."""
from cattrs.preconf.json import configure_converter, JsonConverter, make_converter

__all__ = ["configure_converter", "JsonConverter", "make_converter"]



from attrs import Factory, define

from cattrs import Converter
from cattrs._compat import Final


@define
class C:
    a: Final[int]


def test_unstructure_final(genconverter: Converter) -> None:
    """Unstructuring should work, and unstructure hooks should work."""
    assert genconverter.unstructure(C(1)) == {"a": 1}

    genconverter.register_unstructure_hook(int, lambda i: str(i))
    assert genconverter.unstructure(C(1)) == {"a": "1"}


def test_structure_final(genconverter: Converter) -> None:
    """Structuring should work, and structure hooks should work."""
    assert genconverter.structure({"a": 1}, C) == C(1)

    genconverter.register_structure_hook(int, lambda i, _: int(i) + 1)
    assert genconverter.structure({"a": "1"}, C) == C(2)


@define
class D:
    a: Final[int]
    b: Final = 5
    c: Final = Factory(lambda: 3)


def test_unstructure_bare_final(genconverter: Converter) -> None:
    """Unstructuring bare Finals should work, and unstructure hooks should work."""
    assert genconverter.unstructure(D(1)) == {"a": 1, "b": 5, "c": 3}

    genconverter.register_unstructure_hook(int, lambda i: str(i))
    # Bare finals don't work with factories.
    assert genconverter.unstructure(D(1)) == {"a": "1", "b": "5", "c": 3}


def test_structure_bare_final(genconverter: Converter) -> None:
    """Structuring should work, and structure hooks should work."""
    assert genconverter.structure({"a": 1, "b": 3}, D) == D(1, 3)

    genconverter.register_structure_hook(int, lambda i, _: int(i) + 1)
    assert genconverter.structure({"a": "1", "b": "3"}, D) == D(2, 4, 3)



import typing
from copy import deepcopy
from functools import partial
from typing import Tuple

import attr
import pytest

from cattrs import Converter, override
from cattrs.errors import ClassValidationError
from cattrs.strategies import configure_tagged_union, include_subclasses


@attr.define
class Parent:
    p: int


@attr.define
class Child1(Parent):
    c1: int


@attr.define
class GrandChild(Child1):
    g: int


@attr.define
class Child2(Parent):
    c2: int


@attr.define
class UnionCompose:
    a: typing.Union[Parent, Child1, Child2, GrandChild]


@attr.define
class NonUnionCompose:
    a: Parent


@attr.define
class UnionContainer:
    a: typing.List[typing.Union[Parent, Child1, Child2, GrandChild]]


@attr.define
class NonUnionContainer:
    a: typing.List[Parent]


@attr.define
class CircularA:
    a: int
    other: "typing.List[CircularA]"


@attr.define
class CircularB(CircularA):
    b: int


def _remove_type_name(unstructured: typing.Union[typing.Dict, typing.List]):
    if isinstance(unstructured, list):
        iterator = unstructured
    elif isinstance(unstructured, dict):
        if "type_name" in unstructured:
            unstructured.pop("type_name")
        iterator = unstructured.values()
    for item in iterator:
        if isinstance(item, (list, dict)):
            _remove_type_name(item)
    return unstructured


IDS_TO_STRUCT_UNSTRUCT = {
    "parent-only": (Parent(1), dict(p=1, type_name="Parent")),
    "child1-only": (Child1(1, 2), dict(p=1, c1=2, type_name="Child1")),
    "child2-only": (Child2(1, 2), dict(p=1, c2=2, type_name="Child2")),
    "grandchild-only": (
        GrandChild(1, 2, 3),
        dict(p=1, c1=2, g=3, type_name="GrandChild"),
    ),
    "union-compose-parent": (
        UnionCompose(Parent(1)),
        dict(a=dict(p=1, type_name="Parent")),
    ),
    "union-compose-child": (
        UnionCompose(Child1(1, 2)),
        dict(a=dict(p=1, c1=2, type_name="Child1")),
    ),
    "union-compose-grandchild": (
        UnionCompose(GrandChild(1, 2, 3)),
        dict(a=(dict(p=1, c1=2, g=3, type_name="GrandChild"))),
    ),
    "non-union-compose-parent": (
        NonUnionCompose(Parent(1)),
        dict(a=dict(p=1, type_name="Parent")),
    ),
    "non-union-compose-child": (
        NonUnionCompose(Child1(1, 2)),
        dict(a=dict(p=1, c1=2, type_name="Child1")),
    ),
    "non-union-compose-grandchild": (
        NonUnionCompose(GrandChild(1, 2, 3)),
        dict(a=(dict(p=1, c1=2, g=3, type_name="GrandChild"))),
    ),
    "union-container": (
        UnionContainer([Parent(1), GrandChild(1, 2, 3)]),
        dict(
            a=[
                dict(p=1, type_name="Parent"),
                dict(p=1, c1=2, g=3, type_name="GrandChild"),
            ]
        ),
    ),
    "non-union-container": (
        NonUnionContainer([Parent(1), GrandChild(1, 2, 3)]),
        dict(
            a=[
                dict(p=1, type_name="Parent"),
                dict(p=1, c1=2, g=3, type_name="GrandChild"),
            ]
        ),
    ),
}


@pytest.fixture(
    params=["with-subclasses", "with-subclasses-and-tagged-union", "wo-subclasses"]
)
def conv_w_subclasses(request):
    c = Converter()
    if request.param == "with-subclasses":
        include_subclasses(Parent, c)
        include_subclasses(CircularA, c)
    elif request.param == "with-subclasses-and-tagged-union":
        union_strategy = partial(configure_tagged_union, tag_name="type_name")
        include_subclasses(Parent, c, union_strategy=union_strategy)
        include_subclasses(CircularA, c, union_strategy=union_strategy)

    return c, request.param


@pytest.mark.parametrize(
    "struct_unstruct", IDS_TO_STRUCT_UNSTRUCT.values(), ids=IDS_TO_STRUCT_UNSTRUCT
)
def test_structuring_with_inheritance(
    conv_w_subclasses: Tuple[Converter, bool], struct_unstruct
) -> None:
    structured, unstructured = struct_unstruct

    converter, included_subclasses_param = conv_w_subclasses
    if included_subclasses_param != "with-subclasses-and-tagged-union":
        unstructured = _remove_type_name(deepcopy(unstructured))

    if "wo-subclasses" in included_subclasses_param and isinstance(
        structured, (NonUnionContainer, NonUnionCompose)
    ):
        pytest.xfail(
            "Cannot structure subclasses if include_subclasses strategy is not used"
        )
    assert converter.structure(unstructured, structured.__class__) == structured

    if structured.__class__ in {Child1, Child2, GrandChild}:
        if "wo-subclasses" in included_subclasses_param:
            pytest.xfail(
                "Cannot structure subclasses if include_subclasses strategy is not used"
            )
        assert converter.structure(unstructured, Parent) == structured

    if structured.__class__ == GrandChild:
        assert converter.structure(unstructured, Child1) == structured

    if structured.__class__ in {Parent, Child1, Child2}:
        with pytest.raises(ClassValidationError):
            _ = converter.structure(unstructured, GrandChild)


def test_structure_as_union():
    converter = Converter()
    include_subclasses(Parent, converter)
    the_list = [dict(p=1, c1=2)]
    res = converter.structure(the_list, typing.List[typing.Union[Parent, Child1]])
    assert res == [Child1(1, 2)]


def test_circular_reference(conv_w_subclasses):
    c, included_subclasses_param = conv_w_subclasses

    struct = CircularB(a=1, other=[CircularB(a=2, other=[], b=3)], b=4)
    unstruct = dict(
        a=1,
        other=[dict(a=2, other=[], b=3, type_name="CircularB")],
        b=4,
        type_name="CircularB",
    )

    if included_subclasses_param != "with-subclasses-and-tagged-union":
        unstruct = _remove_type_name(unstruct)

    if "wo-subclasses" in included_subclasses_param:
        pytest.xfail("Cannot succeed if include_subclasses strategy is not used")

    res = c.unstructure(struct)

    assert res == unstruct

    res = c.unstructure(struct, CircularA)
    assert res == unstruct

    res = c.structure(unstruct, CircularA)
    assert res == struct


@pytest.mark.parametrize(
    "struct_unstruct", IDS_TO_STRUCT_UNSTRUCT.values(), ids=IDS_TO_STRUCT_UNSTRUCT
)
def test_unstructuring_with_inheritance(
    conv_w_subclasses: Tuple[Converter, bool], struct_unstruct
):
    structured, unstructured = struct_unstruct
    converter, included_subclasses_param = conv_w_subclasses

    if "wo-subclasses" in included_subclasses_param:
        if isinstance(structured, (NonUnionContainer, NonUnionCompose)):
            pytest.xfail("Cannot succeed if include_subclasses strategy is not used")

    if included_subclasses_param != "with-subclasses-and-tagged-union":
        unstructured = _remove_type_name(deepcopy(unstructured))

    assert converter.unstructure(structured) == unstructured

    if structured.__class__ in {Child1, Child2, GrandChild}:
        if "wo-subclasses" in included_subclasses_param:
            pytest.xfail("Cannot succeed if include_subclasses strategy is not used")
        assert converter.unstructure(structured, unstructure_as=Parent) == unstructured

    if structured.__class__ == GrandChild:
        assert converter.unstructure(structured, unstructure_as=Child1) == unstructured


def test_structuring_unstructuring_unknown_subclass():
    @attr.define
    class A:
        a: int

    @attr.define
    class A1(A):
        a1: int

    converter = Converter()
    include_subclasses(A, converter)

    # We define A2 after having created the custom un/structuring functions for A and A1
    @attr.define
    class A2(A1):
        a2: int

    # Even if A2 did not exist, unstructuring_as A works:
    assert converter.unstructure(A2(1, 2, 3), unstructure_as=A) == dict(a=1, a1=2, a2=3)

    # As well as when unstructuring as A1, in other words, unstructuring works for
    # unknown classes.
    assert converter.unstructure(A2(1, 2, 3), unstructure_as=A1) == {
        "a": 1,
        "a1": 2,
        "a2": 3,
    }

    # But as expected, structuring unknown classes as their parent fails to give the
    # correct answer. This is a documented drawback, we just confirm it.
    assert converter.structure(dict(a=1, a1=2, a2=3), A) == A1(1, 2)


def test_structuring_with_subclasses_argument():
    c = Converter()
    include_subclasses(Parent, c, subclasses=(Child1,))

    structured_child, unstructured_child = IDS_TO_STRUCT_UNSTRUCT[
        "non-union-compose-child"
    ]
    unstructured_child = _remove_type_name(deepcopy(unstructured_child))
    assert c.structure(unstructured_child, NonUnionCompose) == structured_child
    assert c.unstructure(structured_child) == unstructured_child

    structured_gchild, unstructured_gchild = IDS_TO_STRUCT_UNSTRUCT[
        "non-union-compose-grandchild"
    ]
    unstructured_gchild = _remove_type_name(deepcopy(unstructured_gchild))
    assert c.structure(unstructured_gchild, NonUnionCompose) == structured_child
    assert c.unstructure(structured_gchild) == unstructured_gchild


@pytest.mark.parametrize(
    "struct_unstruct", ["parent-only", "child1-only", "child2-only", "grandchild-only"]
)
@pytest.mark.parametrize(
    "with_union_strategy",
    [True, False],
    ids=["with-union-strategy", "wo-union-strategy"],
)
def test_overrides(with_union_strategy: bool, struct_unstruct: str):
    c = Converter()
    union_strategy = (
        partial(configure_tagged_union, tag_name="type_name")
        if with_union_strategy
        else None
    )
    include_subclasses(
        Parent, c, overrides={"p": override(rename="u")}, union_strategy=union_strategy
    )

    structured, unstructured = IDS_TO_STRUCT_UNSTRUCT[struct_unstruct]
    unstructured = unstructured.copy()
    val = unstructured.pop("p")
    unstructured["u"] = val
    if not with_union_strategy:
        unstructured.pop("type_name")

    assert c.unstructure(structured) == unstructured
    assert c.structure(unstructured, Parent) == structured
    assert c.structure(unstructured, structured.__class__) == structured



from typing import Type, Union

import attr
import pytest

from cattrs._compat import is_py310_plus
from cattrs.converters import BaseConverter, Converter


@pytest.mark.parametrize("cls", (BaseConverter, Converter))
def test_custom_union_toplevel_roundtrip(cls: Type[BaseConverter]):
    """
    Test custom code union handling.

    We override union unstructuring to add the class type, and union structuring
    to use the class type.
    """
    c = cls()

    @attr.define
    class A:
        a: int

    @attr.define
    class B:
        a: int

    c.register_unstructure_hook(
        Union[A, B], lambda o: {"_type": o.__class__.__name__, **c.unstructure(o)}
    )
    c.register_structure_hook(
        Union[A, B], lambda o, t: c.structure(o, A if o["_type"] == "A" else B)
    )

    inst = B(1)
    unstructured = c.unstructure(inst, unstructure_as=Union[A, B])
    assert unstructured["_type"] == "B"

    assert c.structure(unstructured, Union[A, B]) == inst


@pytest.mark.skipif(not is_py310_plus, reason="3.10 union syntax")
@pytest.mark.parametrize("cls", (BaseConverter, Converter))
def test_310_custom_union_toplevel_roundtrip(cls: Type[BaseConverter]):
    """
    Test custom code union handling.

    We override union unstructuring to add the class type, and union structuring
    to use the class type.
    """
    c = cls()

    @attr.define
    class A:
        a: int

    @attr.define
    class B:
        a: int

    c.register_unstructure_hook(
        A | B, lambda o: {"_type": o.__class__.__name__, **c.unstructure(o)}
    )
    c.register_structure_hook(
        A | B, lambda o, t: c.structure(o, A if o["_type"] == "A" else B)
    )

    inst = B(1)
    unstructured = c.unstructure(inst, unstructure_as=A | B)
    assert unstructured["_type"] == "B"

    assert c.structure(unstructured, A | B) == inst


@pytest.mark.parametrize("cls", (BaseConverter, Converter))
def test_custom_union_clsfield_roundtrip(cls: Type[BaseConverter]):
    """
    Test custom code union handling.

    We override union unstructuring to add the class type, and union structuring
    to use the class type.
    """
    c = cls()

    @attr.define
    class A:
        a: int

    @attr.define
    class B:
        a: int

    @attr.define
    class C:
        f: Union[A, B]

    c.register_unstructure_hook(
        Union[A, B], lambda o: {"_type": o.__class__.__name__, **c.unstructure(o)}
    )
    c.register_structure_hook(
        Union[A, B], lambda o, t: c.structure(o, A if o["_type"] == "A" else B)
    )

    inst = C(A(1))
    unstructured = c.unstructure(inst)
    assert unstructured["f"]["_type"] == "A"

    assert c.structure(unstructured, C) == inst



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Optional

from attr import NOTHING, Attribute, Factory

from .._compat import is_bare_final

if TYPE_CHECKING:  # pragma: no cover
    from cattr.converters import BaseConverter


def find_structure_handler(
    a: Attribute, type: Any, c: BaseConverter, prefer_attrs_converters: bool = False
) -> Optional[Callable[[Any, Any], Any]]:
    """Find the appropriate structure handler to use.

    Return `None` if no handler should be used.
    """
    if a.converter is not None and prefer_attrs_converters:
        # If the user as requested to use attrib converters, use nothing
        # so it falls back to that.
        handler = None
    elif a.converter is not None and not prefer_attrs_converters and type is not None:
        handler = c._structure_func.dispatch(type)
        if handler == c._structure_error:
            handler = None
    elif type is not None:
        if (
            is_bare_final(type)
            and a.default is not NOTHING
            and not isinstance(a.default, Factory)
        ):
            # This is a special case where we can use the
            # type of the default to dispatch on.
            type = a.default.__class__
            handler = c._structure_func.dispatch(type)
            if handler == c._structure_call:
                # Finals can't really be used with _structure_call, so
                # we wrap it so the rest of the toolchain doesn't get
                # confused.

                def handler(v, _, _h=handler):
                    return _h(v, type)

        else:
            handler = c._structure_func.dispatch(type)
    else:
        handler = c.structure
    return handler



"""Strategies for attributes with types and classes using them."""
import sys
from collections import OrderedDict
from collections.abc import MutableSequence as AbcMutableSequence
from collections.abc import MutableSet as AbcMutableSet
from collections.abc import Sequence as AbcSequence
from collections.abc import Set as AbcSet
from dataclasses import field, make_dataclass
from functools import partial
from pathlib import Path
from typing import (
    Any,
    Dict,
    FrozenSet,
    List,
    MutableSequence,
    MutableSet,
    NewType,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
)

import attr
from attr import NOTHING, Factory, frozen
from attr._make import _CountingAttr
from hypothesis import note
from hypothesis.strategies import (
    DrawFn,
    SearchStrategy,
    booleans,
    composite,
    dictionaries,
    fixed_dictionaries,
    floats,
    frozensets,
    integers,
    just,
    lists,
    recursive,
    sampled_from,
    sets,
    text,
    tuples,
)

from .untyped import gen_attr_names, make_class

is_39_or_later = sys.version_info[:2] >= (3, 9)
PosArg = Any
PosArgs = Tuple[PosArg]
KwArgs = Dict[str, Any]
T = TypeVar("T")


def simple_typed_classes(
    defaults=None, min_attrs=0, frozen=False, kw_only=None, newtypes=True
) -> SearchStrategy[Tuple[Type, PosArgs, KwArgs]]:
    """Yield tuples of (class, values)."""
    return lists_of_typed_attrs(
        defaults,
        min_size=min_attrs,
        for_frozen=frozen,
        kw_only=kw_only,
        newtypes=newtypes,
    ).flatmap(partial(_create_hyp_class, frozen=frozen))


def simple_typed_dataclasses(defaults=None, min_attrs=0, frozen=False, newtypes=True):
    """Yield tuples of (class, values)."""
    return lists_of_typed_attrs(
        defaults,
        min_size=min_attrs,
        for_frozen=frozen,
        allow_mutable_defaults=False,
        newtypes=newtypes,
    ).flatmap(partial(_create_dataclass, frozen=frozen))


def simple_typed_classes_and_strats(
    defaults=None, min_attrs=0, kw_only=None, newtypes=True
) -> SearchStrategy[Tuple[Type, SearchStrategy[PosArgs], SearchStrategy[KwArgs]]]:
    """Yield tuples of (class, (strategies))."""
    return lists_of_typed_attrs(
        defaults, min_size=min_attrs, kw_only=kw_only, newtypes=newtypes
    ).flatmap(_create_hyp_class_and_strat)


def lists_of_typed_attrs(
    defaults=None,
    min_size=0,
    for_frozen=False,
    allow_mutable_defaults=True,
    kw_only=None,
    newtypes=True,
) -> SearchStrategy[List[Tuple[_CountingAttr, SearchStrategy[PosArg]]]]:
    # Python functions support up to 255 arguments.
    return lists(
        simple_typed_attrs(
            defaults,
            for_frozen=for_frozen,
            allow_mutable_defaults=allow_mutable_defaults,
            kw_only=kw_only,
            newtypes=newtypes,
        ),
        min_size=min_size,
        max_size=50,
    ).map(
        lambda l: sorted(l, key=lambda t: (t[0]._default is not NOTHING, t[0].kw_only))
    )


def simple_typed_attrs(
    defaults=None,
    for_frozen=False,
    allow_mutable_defaults=True,
    kw_only=None,
    newtypes=True,
) -> SearchStrategy[Tuple[_CountingAttr, SearchStrategy[PosArgs]]]:
    if not is_39_or_later:
        res = (
            bare_typed_attrs(defaults, kw_only)
            | int_typed_attrs(defaults, kw_only)
            | str_typed_attrs(defaults, kw_only)
            | float_typed_attrs(defaults, kw_only)
            | frozenset_typed_attrs(defaults, legacy_types_only=True, kw_only=kw_only)
            | homo_tuple_typed_attrs(defaults, legacy_types_only=True, kw_only=kw_only)
            | path_typed_attrs(defaults, kw_only=kw_only)
        )
        if newtypes:
            res = (
                res
                | newtype_int_typed_attrs(defaults, kw_only)
                | newtype_attrs_typed_attrs(defaults, kw_only)
            )
        if not for_frozen:
            res = (
                res
                | dict_typed_attrs(defaults, allow_mutable_defaults, kw_only)
                | mutable_seq_typed_attrs(
                    defaults,
                    allow_mutable_defaults,
                    legacy_types_only=True,
                    kw_only=kw_only,
                )
                | seq_typed_attrs(
                    defaults,
                    allow_mutable_defaults,
                    legacy_types_only=True,
                    kw_only=kw_only,
                )
                | list_typed_attrs(
                    defaults,
                    allow_mutable_defaults,
                    legacy_types_only=True,
                    kw_only=kw_only,
                )
                | set_typed_attrs(
                    defaults,
                    allow_mutable_defaults,
                    legacy_types_only=True,
                    kw_only=kw_only,
                )
            )
    else:
        res = (
            bare_typed_attrs(defaults, kw_only)
            | int_typed_attrs(defaults, kw_only)
            | str_typed_attrs(defaults, kw_only)
            | float_typed_attrs(defaults, kw_only)
            | frozenset_typed_attrs(defaults, kw_only=kw_only)
            | homo_tuple_typed_attrs(defaults, kw_only=kw_only)
            | path_typed_attrs(defaults, kw_only=kw_only)
        )
        if newtypes:
            res = (
                res
                | newtype_int_typed_attrs(defaults, kw_only)
                | newtype_attrs_typed_attrs(defaults, kw_only)
            )

        if not for_frozen:
            res = (
                res
                | dict_typed_attrs(defaults, allow_mutable_defaults, kw_only)
                | new_dict_typed_attrs(defaults, allow_mutable_defaults, kw_only)
                | set_typed_attrs(defaults, allow_mutable_defaults, kw_only=kw_only)
                | list_typed_attrs(defaults, allow_mutable_defaults, kw_only=kw_only)
                | mutable_seq_typed_attrs(
                    defaults, allow_mutable_defaults, kw_only=kw_only
                )
                | seq_typed_attrs(defaults, allow_mutable_defaults, kw_only=kw_only)
            )

    return res


def _create_hyp_class(
    attrs_and_strategy: List[Tuple[_CountingAttr, SearchStrategy[PosArgs]]],
    frozen=False,
) -> SearchStrategy[Tuple[Type, PosArgs, KwArgs]]:
    """
    A helper function for Hypothesis to generate attrs classes.

    The result is a tuple: an attrs class, tuple of values to
    instantiate it, and a kwargs dict for kw_only args.
    """

    def key(t):
        return (t[0]._default is not attr.NOTHING, t[0].kw_only)

    attrs_and_strat = sorted(attrs_and_strategy, key=key)
    attrs = [a[0] for a in attrs_and_strat]
    for i, a in enumerate(attrs):
        a.counter = i
    vals = tuple((a[1]) for a in attrs_and_strat if not a[0].kw_only)
    note(f"Class fields: {attrs}")
    attrs_dict = OrderedDict(zip(gen_attr_names(), attrs))
    kwarg_strats = {}
    for attr_name, attr_and_strat in zip(gen_attr_names(), attrs_and_strat):
        if attr_and_strat[0].kw_only:
            if attr_name.startswith("_"):
                attr_name = attr_name[1:]
            kwarg_strats[attr_name] = attr_and_strat[1]

    return tuples(
        just(make_class("HypAttrsClass", attrs_dict, frozen=frozen)),
        tuples(*vals),
        fixed_dictionaries(kwarg_strats),
    )


def _create_dataclass(
    attrs_and_strategy: List[Tuple[_CountingAttr, SearchStrategy[PosArgs]]],
    frozen=False,
) -> SearchStrategy[Tuple[Type, PosArgs, KwArgs]]:
    """
    A helper function for Hypothesis to generate dataclasses.

    The result is a tuple: a dataclass, a tuple of values to
    instantiate it, and an empty dict (usually used for kw-only attrs attributes).
    """

    def key(t):
        return t[0]._default is not attr.NOTHING

    attrs_and_strat = sorted(attrs_and_strategy, key=key)
    attrs = [a[0] for a in attrs_and_strat]
    for i, a in enumerate(attrs):
        a.counter = i
    vals = tuple((a[1]) for a in attrs_and_strat)
    return tuples(
        just(
            make_dataclass(
                "HypDataclass",
                [
                    (n, a.type)
                    if a._default is NOTHING
                    else (
                        (n, a.type, field(default=a._default))
                        if not isinstance(a._default, Factory)
                        else (n, a.type, field(default_factory=a._default.factory))
                    )
                    for n, a in zip(gen_attr_names(), attrs)
                ],
                frozen=frozen,
            )
        ),
        tuples(*vals),
        just({}),
    )


def _create_hyp_class_and_strat(
    attrs_and_strategy: List[Tuple[_CountingAttr, SearchStrategy[PosArg]]]
) -> SearchStrategy[Tuple[Type, SearchStrategy[PosArgs], SearchStrategy[KwArgs]]]:
    def key(t):
        return (t[0].default is not attr.NOTHING, t[0].kw_only)

    attrs_and_strat = sorted(attrs_and_strategy, key=key)
    attrs = [a[0] for a in attrs_and_strat]
    for i, a in enumerate(attrs):
        a.counter = i
    vals = tuple((a[1]) for a in attrs_and_strat if not a[0].kw_only)
    kwarg_strats = {}
    for attr_name, attr_and_strat in zip(gen_attr_names(), attrs_and_strat):
        if attr_and_strat[0].kw_only:
            if attr_name.startswith("_"):
                attr_name = attr_name[1:]
            kwarg_strats[attr_name] = attr_and_strat[1]
    return tuples(
        just(make_class("HypClass", OrderedDict(zip(gen_attr_names(), attrs)))),
        just(tuples(*vals)),
        just(fixed_dictionaries(kwarg_strats)),
    )


@composite
def bare_typed_attrs(draw, defaults=None, kw_only=None):
    """
    Generate a tuple of an attribute and a strategy that yields values
    appropriate for that attribute.
    """
    default = attr.NOTHING
    if defaults is True or (defaults is None and draw(booleans())):
        default = None
    return (
        attr.ib(
            type=Any,
            default=default,
            kw_only=draw(booleans()) if kw_only is None else kw_only,
        ),
        just(None),
    )


@composite
def int_typed_attrs(draw, defaults=None, kw_only=None):
    """
    Generate a tuple of an attribute and a strategy that yields ints for that
    attribute.
    """
    default = attr.NOTHING
    if defaults is True or (defaults is None and draw(booleans())):
        default = draw(integers())
    return (
        attr.ib(
            type=int,
            default=default,
            kw_only=draw(booleans()) if kw_only is None else kw_only,
        ),
        integers(),
    )


@composite
def str_typed_attrs(draw, defaults=None, kw_only=None):
    """
    Generate a tuple of an attribute and a strategy that yields strs for that
    attribute.
    """
    default = NOTHING
    if defaults is True or (defaults is None and draw(booleans())):
        default = draw(text())
    return (
        attr.ib(
            type=str,
            default=default,
            kw_only=draw(booleans()) if kw_only is None else kw_only,
        ),
        text(),
    )


@composite
def float_typed_attrs(draw, defaults=None, kw_only=None):
    """
    Generate a tuple of an attribute and a strategy that yields floats for that
    attribute.
    """
    default = attr.NOTHING
    if defaults is True or (defaults is None and draw(booleans())):
        default = draw(floats())
    return (
        attr.ib(
            type=float,
            default=default,
            kw_only=draw(booleans()) if kw_only is None else kw_only,
        ),
        floats(),
    )


@composite
def path_typed_attrs(
    draw: DrawFn, defaults: Optional[bool] = None, kw_only: Optional[bool] = None
) -> Tuple[_CountingAttr, SearchStrategy[Path]]:
    """
    Generate a tuple of an attribute and a strategy that yields paths for that
    attribute.
    """
    from string import ascii_lowercase

    default = attr.NOTHING
    if defaults is True or (defaults is None and draw(booleans())):
        default = Path(draw(text(ascii_lowercase, min_size=1)))
    return (
        attr.ib(
            type=Path,
            default=default,
            kw_only=draw(booleans()) if kw_only is None else kw_only,
        ),
        text(ascii_lowercase, min_size=1).map(Path),
    )


@composite
def dict_typed_attrs(
    draw, defaults=None, allow_mutable_defaults=True, kw_only=None
) -> SearchStrategy[Tuple[_CountingAttr, SearchStrategy]]:
    """
    Generate a tuple of an attribute and a strategy that yields dictionaries
    for that attribute. The dictionaries map strings to integers.
    The generated dict types are what's expected to be used on pre-3.9 Pythons.
    """
    default = attr.NOTHING
    val_strat = dictionaries(keys=text(), values=integers())
    if defaults is True or (defaults is None and draw(booleans())):
        default_val = draw(val_strat)
        if not allow_mutable_defaults or draw(booleans()):
            default = Factory(lambda: default_val)
        else:
            default = default_val
    type = draw(sampled_from([Dict[str, int], Dict, dict]))
    return (
        attr.ib(
            type=type,
            default=default,
            kw_only=draw(booleans()) if kw_only is None else kw_only,
        ),
        val_strat,
    )


@composite
def new_dict_typed_attrs(
    draw, defaults=None, allow_mutable_defaults=True, kw_only=None
):
    """
    Generate a tuple of an attribute and a strategy that yields dictionaries
    for that attribute. The dictionaries map strings to integers.

    Uses the new 3.9 dict annotation.
    """
    default_val = attr.NOTHING
    val_strat = dictionaries(keys=text(), values=integers())
    if defaults is True or (defaults is None and draw(booleans())):
        default_val = draw(val_strat)
        if not allow_mutable_defaults or draw(booleans()):
            default = Factory(lambda: default_val)
        else:
            default = default_val
    else:
        default = default_val

    return (
        attr.ib(
            type=dict[str, int],
            default=default,
            kw_only=draw(booleans()) if kw_only is None else kw_only,
        ),
        val_strat,
    )


@composite
def set_typed_attrs(
    draw: DrawFn,
    defaults=None,
    allow_mutable_defaults=True,
    legacy_types_only=False,
    kw_only=None,
):
    """
    Generate a tuple of an attribute and a strategy that yields sets
    for that attribute. The sets contain integers.
    """
    default_val = attr.NOTHING
    val_strat = sets(integers())
    if defaults is True or (defaults is None and draw(booleans())):
        default_val = draw(val_strat)
        if not allow_mutable_defaults or draw(booleans()):
            default = Factory(lambda: default_val)
        else:
            default = default_val
    else:
        default = default_val

    type = draw(
        sampled_from(
            [set, set[int], AbcSet[int], AbcMutableSet[int]]
            if not legacy_types_only
            else [set, Set[int], MutableSet[int]]
        )
    )
    return (
        attr.ib(
            type=type,
            default=default,
            kw_only=draw(booleans()) if kw_only is None else kw_only,
        ),
        val_strat,
    )


@composite
def frozenset_typed_attrs(
    draw: DrawFn, defaults=None, legacy_types_only=False, kw_only=None
):
    """
    Generate a tuple of an attribute and a strategy that yields frozensets
    for that attribute. The frozensets contain integers.
    """
    default = attr.NOTHING
    val_strat = frozensets(integers())
    if defaults is True or (defaults is None and draw(booleans())):
        default = draw(val_strat)
    type = draw(
        sampled_from(
            [frozenset[int], frozenset, FrozenSet[int], FrozenSet]
            if not legacy_types_only
            else [frozenset, FrozenSet[int], FrozenSet]
        )
    )
    return (
        attr.ib(
            type=type,
            default=default,
            kw_only=draw(booleans()) if kw_only is None else kw_only,
        ),
        val_strat,
    )


@composite
def list_typed_attrs(
    draw: DrawFn,
    defaults=None,
    allow_mutable_defaults=True,
    legacy_types_only=False,
    kw_only=None,
):
    """
    Generate a tuple of an attribute and a strategy that yields lists
    for that attribute. The lists contain floats.
    """
    default_val = attr.NOTHING
    val_strat = lists(floats(allow_infinity=False, allow_nan=False))
    if defaults is True or (defaults is None and draw(booleans())):
        default_val = draw(val_strat)
        if not allow_mutable_defaults or draw(booleans()):
            default = Factory(lambda: default_val)
        else:
            default = default_val
    else:
        default = default_val
    return (
        attr.ib(
            type=draw(
                sampled_from(
                    [list[float], list, List[float], List]
                    if not legacy_types_only
                    else [List, List[float], list]
                )
            ),
            default=default,
            kw_only=draw(booleans()) if kw_only is None else kw_only,
        ),
        val_strat,
    )


@composite
def seq_typed_attrs(
    draw,
    defaults=None,
    allow_mutable_defaults=True,
    legacy_types_only=False,
    kw_only=None,
):
    """
    Generate a tuple of an attribute and a strategy that yields lists
    for that attribute. The lists contain integers.
    """
    default_val = attr.NOTHING
    val_strat = lists(integers())
    if defaults is True or (defaults is None and draw(booleans())):
        default_val = draw(val_strat)
        if not allow_mutable_defaults or draw(booleans()):
            default = Factory(lambda: default_val)
        else:
            default = default_val
    else:
        default = default_val

    return (
        attr.ib(
            type=AbcSequence[int] if not legacy_types_only else Sequence[int],
            default=default,
            kw_only=draw(booleans()) if kw_only is None else kw_only,
        ),
        val_strat,
    )


@composite
def mutable_seq_typed_attrs(
    draw,
    defaults=None,
    allow_mutable_defaults=True,
    legacy_types_only=False,
    kw_only=None,
):
    """
    Generate a tuple of an attribute and a strategy that yields lists
    for that attribute. The lists contain floats.
    """
    default_val = attr.NOTHING
    val_strat = lists(floats(allow_infinity=False, allow_nan=False))
    if defaults is True or (defaults is None and draw(booleans())):
        default_val = draw(val_strat)
        if not allow_mutable_defaults or draw(booleans()):
            default = Factory(lambda: default_val)
        else:
            default = default_val
    else:
        default = default_val

    return (
        attr.ib(
            type=AbcMutableSequence[float]
            if not legacy_types_only
            else MutableSequence[float],
            default=default,
            kw_only=draw(booleans()) if kw_only is None else kw_only,
        ),
        val_strat,
    )


@composite
def homo_tuple_typed_attrs(draw, defaults=None, legacy_types_only=False, kw_only=None):
    """
    Generate a tuple of an attribute and a strategy that yields homogenous
    tuples for that attribute. The tuples contain strings.
    """
    default = attr.NOTHING
    val_strat = tuples(text(), text(), text())
    if defaults is True or (defaults is None and draw(booleans())):
        default = draw(val_strat)
    return (
        attr.ib(
            type=draw(
                sampled_from(
                    [tuple[str, ...], tuple, Tuple, Tuple[str, ...]]
                    if not legacy_types_only
                    else [tuple, Tuple, Tuple[str, ...]]
                )
            ),
            default=default,
            kw_only=draw(booleans()) if kw_only is None else kw_only,
        ),
        val_strat,
    )


@composite
def newtype_int_typed_attrs(draw: DrawFn, defaults=None, kw_only=None):
    """
    Generate a tuple of an attribute and a strategy that yields ints for that
    attribute.
    """
    default = attr.NOTHING
    if defaults is True or (defaults is None and draw(booleans())):
        default = draw(integers())
    type = NewType("NewInt", int)
    return (
        attr.ib(
            type=type,
            default=default,
            kw_only=draw(booleans()) if kw_only is None else kw_only,
        ),
        integers(),
    )


@composite
def newtype_attrs_typed_attrs(draw: DrawFn, defaults=None, kw_only=None):
    """
    Generate a tuple of an attribute and a strategy that yields values for that
    attribute.
    """
    default = attr.NOTHING

    @frozen
    class NewTypeAttrs:
        a: int

    if defaults is True or (defaults is None and draw(booleans())):
        default = NewTypeAttrs(draw(integers()))

    type = NewType("NewAttrs", NewTypeAttrs)
    return (
        attr.ib(
            type=type,
            default=default,
            kw_only=draw(booleans()) if kw_only is None else kw_only,
        ),
        integers().map(NewTypeAttrs),
    )


def just_class(
    tup: Tuple[
        List[Tuple[_CountingAttr, SearchStrategy]], Tuple[Type, PosArgs, KwArgs]
    ],
    defaults: Tuple[PosArgs, KwArgs],
):
    nested_cl = tup[1][0]
    nested_cl_args = tup[1][1]
    nested_cl_kwargs = tup[1][2]
    default = attr.Factory(lambda: nested_cl(*defaults[0], **defaults[1]))
    combined_attrs = list(tup[0])
    combined_attrs.append(
        (
            attr.ib(type=nested_cl, default=default),
            just(nested_cl(*nested_cl_args, **nested_cl_kwargs)),
        )
    )
    return _create_hyp_class_and_strat(combined_attrs)


def list_of_class(
    tup: Tuple[
        List[Tuple[_CountingAttr, SearchStrategy]], Tuple[Type, PosArgs, KwArgs]
    ],
    defaults: Tuple[PosArgs, KwArgs],
) -> SearchStrategy[Tuple[Type, SearchStrategy[PosArgs], SearchStrategy[KwArgs]]]:
    nested_cl = tup[1][0]
    nested_cl_args = tup[1][1]
    nested_cl_kwargs = tup[1][2]
    default = attr.Factory(lambda: [nested_cl(*defaults[0], **defaults[1])])
    combined_attrs = list(tup[0])
    combined_attrs.append(
        (
            attr.ib(type=List[nested_cl], default=default),
            just([nested_cl(*nested_cl_args, **nested_cl_kwargs)]),
        )
    )
    return _create_hyp_class_and_strat(combined_attrs)


def new_list_of_class(
    tup: Tuple[
        List[Tuple[_CountingAttr, SearchStrategy]], Tuple[Type, PosArgs, KwArgs]
    ],
    defaults: Tuple[PosArgs, KwArgs],
):
    """Uses the new 3.9 list type annotation."""
    nested_cl = tup[1][0]
    nested_cl_args = tup[1][1]
    nested_cl_kwargs = tup[1][2]
    default = attr.Factory(lambda: [nested_cl(*defaults[0], **defaults[1])])
    combined_attrs = list(tup[0])
    combined_attrs.append(
        (
            attr.ib(type=list[nested_cl], default=default),
            just([nested_cl(*nested_cl_args, **nested_cl_kwargs)]),
        )
    )
    return _create_hyp_class_and_strat(combined_attrs)


def dict_of_class(
    tup: Tuple[
        List[Tuple[_CountingAttr, SearchStrategy]], Tuple[Type, PosArgs, KwArgs]
    ],
    defaults: Tuple[PosArgs, KwArgs],
):
    nested_cl = tup[1][0]
    nested_cl_args = tup[1][1]
    nested_cl_kwargs = tup[1][2]
    default = attr.Factory(lambda: {"cls": nested_cl(*defaults[0], **defaults[1])})
    combined_attrs = list(tup[0])
    combined_attrs.append(
        (
            attr.ib(type=Dict[str, nested_cl], default=default),
            just({"cls": nested_cl(*nested_cl_args, **nested_cl_kwargs)}),
        )
    )
    return _create_hyp_class_and_strat(combined_attrs)


def _create_hyp_nested_strategy(
    simple_class_strategy: SearchStrategy, kw_only=None, newtypes=True
) -> SearchStrategy[Tuple[Type, SearchStrategy[PosArgs], SearchStrategy[KwArgs]]]:
    """
    Create a recursive attrs class.
    Given a strategy for building (simpler) classes, create and return
    a strategy for building classes that have as an attribute:
        * just the simpler class
        * a list of simpler classes
        * a dict mapping the string "cls" to a simpler class.
    """
    # A strategy producing tuples of the form ([list of attributes], <given
    # class strategy>).
    attrs_and_classes: SearchStrategy[
        Tuple[List[Tuple[_CountingAttr, PosArgs]], Tuple[Type, SearchStrategy[PosArgs]]]
    ] = tuples(
        lists_of_typed_attrs(kw_only=kw_only, newtypes=newtypes), simple_class_strategy
    )

    return nested_classes(attrs_and_classes)


@composite
def nested_classes(
    draw: DrawFn,
    attrs_and_classes: SearchStrategy[
        Tuple[
            List[Tuple[_CountingAttr, SearchStrategy]],
            Tuple[Type, SearchStrategy[PosArgs], SearchStrategy[KwArgs]],
        ]
    ],
) -> SearchStrategy[Tuple[Type, SearchStrategy[PosArgs], SearchStrategy[KwArgs]]]:
    attrs, class_and_strat = draw(attrs_and_classes)
    cls, strat, kw_strat = class_and_strat
    pos_defs = tuple(draw(strat))
    kwarg_defs = draw(kw_strat)
    init_vals = tuple(draw(strat))
    init_kwargs = draw(kw_strat)
    if is_39_or_later:
        return draw(
            list_of_class(
                (attrs, (cls, init_vals, init_kwargs)), (pos_defs, kwarg_defs)
            )
            | new_list_of_class(
                (attrs, (cls, init_vals, init_kwargs)), (pos_defs, kwarg_defs)
            )
            | dict_of_class(
                (attrs, (cls, init_vals, init_kwargs)), (pos_defs, kwarg_defs)
            )
            | just_class((attrs, (cls, init_vals, init_kwargs)), (pos_defs, kwarg_defs))
        )
    else:
        return draw(
            list_of_class(
                (attrs, (cls, init_vals, init_kwargs)), (pos_defs, kwarg_defs)
            )
            | dict_of_class(
                (attrs, (cls, init_vals, init_kwargs)), (pos_defs, kwarg_defs)
            )
            | just_class((attrs, (cls, init_vals, init_kwargs)), (pos_defs, kwarg_defs))
        )


def nested_typed_classes_and_strat(
    defaults=None, min_attrs=0, kw_only=None, newtypes=True
) -> SearchStrategy[Tuple[Type, SearchStrategy[PosArgs]]]:
    return recursive(
        simple_typed_classes_and_strats(
            defaults=defaults, min_attrs=min_attrs, kw_only=kw_only, newtypes=newtypes
        ),
        partial(_create_hyp_nested_strategy, kw_only=kw_only, newtypes=newtypes),
        max_leaves=20,
    )


@composite
def nested_typed_classes(draw, defaults=None, min_attrs=0, kw_only=None, newtypes=True):
    cl, strat, kwarg_strat = draw(
        nested_typed_classes_and_strat(
            defaults=defaults, min_attrs=min_attrs, kw_only=kw_only, newtypes=newtypes
        )
    )
    return cl, draw(strat), draw(kwarg_strat)



"""Strategies for attributes without types and accompanying classes."""
import keyword
import string
from collections import OrderedDict
from enum import Enum
from typing import (
    Any,
    Deque,
    Dict,
    Iterable,
    List,
    Mapping,
    MutableMapping,
    MutableSequence,
    MutableSet,
    Sequence,
    Set,
    Tuple,
)

import attr
from attr import NOTHING, make_class
from attr._make import _CountingAttr
from hypothesis import strategies as st

PosArg = Any
PosArgs = Tuple[PosArg]
KwArgs = Dict[str, Any]

primitive_strategies = st.sampled_from(
    [
        (st.integers(), int),
        (st.floats(allow_nan=False), float),
        (st.text(), str),
        (st.binary(), bytes),
    ]
)


@st.composite
def enums_of_primitives(draw):
    """Generate enum classes with primitive values."""
    names = draw(
        st.sets(st.text(min_size=1).filter(lambda s: not s.endswith("_")), min_size=1)
    )
    n = len(names)
    vals = draw(
        st.one_of(
            st.sets(
                st.one_of(
                    st.integers(), st.floats(allow_nan=False), st.text(min_size=1)
                ),
                min_size=n,
                max_size=n,
            )
        )
    )
    return Enum("HypEnum", list(zip(names, vals)))


list_types = st.sampled_from([List, Sequence, MutableSequence])
deque_types = st.sampled_from([Deque, Sequence, MutableSequence])
set_types = st.sampled_from([Set, MutableSet])


@st.composite
def lists_of_primitives(draw):
    """Generate a strategy that yields tuples of list of primitives and types.

    For example, a sample value might be ([1,2], List[int]).
    """
    prim_strat, t = draw(primitive_strategies)
    list_t = draw(list_types.map(lambda list_t: list_t[t]) | list_types)
    return draw(st.lists(prim_strat)), list_t


@st.composite
def deques_of_primitives(draw):
    """Generate a strategy that yields tuples of list of primitives and types.

    For example, a sample value might be ([1,2], Deque[int]).
    """
    prim_strat, t = draw(primitive_strategies)
    deque_t = draw(deque_types.map(lambda deque_t: deque_t[t]) | deque_types)
    return draw(st.lists(prim_strat)), deque_t


@st.composite
def mut_sets_of_primitives(draw):
    """A strategy that generates mutable sets of primitives."""
    prim_strat, t = draw(primitive_strategies)
    set_t = draw(set_types.map(lambda set_t: set_t[t]) | set_types)
    return draw(st.sets(prim_strat)), set_t


@st.composite
def frozen_sets_of_primitives(draw):
    """A strategy that generates frozen sets of primitives."""
    prim_strat, t = draw(primitive_strategies)
    set_t = draw(st.just(Set) | st.just(Set[t]))
    return frozenset(draw(st.sets(prim_strat))), set_t


h_tuple_types = st.sampled_from([Tuple, Sequence])
h_tuples_of_primitives = primitive_strategies.flatmap(
    lambda e: st.tuples(
        st.lists(e[0]),
        st.one_of(st.sampled_from([Tuple[e[1], ...], Sequence[e[1]]]), h_tuple_types),
    )
).map(lambda e: (tuple(e[0]), e[1]))

dict_types = st.sampled_from([Dict, MutableMapping, Mapping])

seqs_of_primitives = st.one_of(lists_of_primitives(), h_tuples_of_primitives)
deque_seqs_of_primitives = st.one_of(deques_of_primitives(), h_tuples_of_primitives)
sets_of_primitives = st.one_of(mut_sets_of_primitives(), frozen_sets_of_primitives())


def create_generic_dict_type(type1, type2):
    """Create a strategy for generating parameterized dict types."""
    return st.one_of(
        dict_types,
        dict_types.map(lambda t: t[type1, type2]),
        dict_types.map(lambda t: t[Any, type2]),
        dict_types.map(lambda t: t[type1, Any]),
    )


def create_dict_and_type(tuple_of_strats):
    """Map two primitive strategies into a strategy for dict and type."""
    (prim_strat_1, type_1), (prim_strat_2, type_2) = tuple_of_strats

    return st.tuples(
        st.dictionaries(prim_strat_1, prim_strat_2),
        create_generic_dict_type(type_1, type_2),
    )


dicts_of_primitives = st.tuples(primitive_strategies, primitive_strategies).flatmap(
    create_dict_and_type
)


def gen_attr_names() -> Iterable[str]:
    """
    Generate names for attributes, 'a'...'z', then 'aa'...'zz'.
    ~702 different attribute names should be enough in practice.
    Some short strings (such as 'as') are keywords, so we skip them.

    Every second attribute name is private (starts with an underscore).
    """
    lc = string.ascii_lowercase
    has_underscore = False
    for c in lc:
        yield c if not has_underscore else "_" + c
        has_underscore = not has_underscore
    for outer in lc:
        for inner in lc:
            res = outer + inner
            if keyword.iskeyword(res):
                continue
            yield outer + inner


def _create_hyp_class(
    attrs_and_strategy: List[Tuple[_CountingAttr, st.SearchStrategy[PosArgs]]],
    frozen=None,
):
    """
    A helper function for Hypothesis to generate attrs classes.

    The result is a tuple: an attrs class, a tuple of values to
    instantiate it, and a kwargs dict for kw-only attributes.
    """

    def key(t):
        return (t[0].default is not NOTHING, t[0].kw_only)

    attrs_and_strat = sorted(attrs_and_strategy, key=key)
    attrs = [a[0] for a in attrs_and_strat]
    for i, a in enumerate(attrs):
        a.counter = i
    vals = tuple((a[1]) for a in attrs_and_strat if not a[0].kw_only)
    kwargs = {}
    for attr_name, attr_and_strat in zip(gen_attr_names(), attrs_and_strat):
        if attr_and_strat[0].kw_only:
            if attr_name.startswith("_"):
                attr_name = attr_name[1:]
            kwargs[attr_name] = attr_and_strat[1]
    return st.tuples(
        st.builds(
            lambda f: make_class(
                "HypClass", OrderedDict(zip(gen_attr_names(), attrs)), frozen=f
            ),
            st.booleans() if frozen is None else st.just(frozen),
        ),
        st.tuples(*vals),
        st.fixed_dictionaries(kwargs),
    )


def just_class(tup):
    nested_cl = tup[1][0]
    default = attr.Factory(nested_cl)
    combined_attrs = list(tup[0])
    combined_attrs.append((attr.ib(default=default), st.just(nested_cl())))
    return _create_hyp_class(combined_attrs)


def just_class_with_type(tup):
    nested_cl = tup[1][0]
    default = attr.Factory(nested_cl)
    combined_attrs = list(tup[0])
    combined_attrs.append(
        (attr.ib(default=default, type=nested_cl), st.just(nested_cl()))
    )
    return _create_hyp_class(combined_attrs)


def just_class_with_type_takes_self(tup):
    nested_cl = tup[1][0]
    default = attr.Factory(lambda _: nested_cl(), takes_self=True)
    combined_attrs = list(tup[0])
    combined_attrs.append(
        (attr.ib(default=default, type=nested_cl), st.just(nested_cl()))
    )
    return _create_hyp_class(combined_attrs)


def just_frozen_class_with_type(tup):
    nested_cl = tup[1][0]
    combined_attrs = list(tup[0])
    combined_attrs.append(
        (attr.ib(default=nested_cl(), type=nested_cl), st.just(nested_cl()))
    )
    return _create_hyp_class(combined_attrs)


def list_of_class(tup):
    nested_cl = tup[1][0]
    default = attr.Factory(lambda: [nested_cl()])
    combined_attrs = list(tup[0])
    combined_attrs.append((attr.ib(default=default), st.just([nested_cl()])))
    return _create_hyp_class(combined_attrs)


def list_of_class_with_type(tup):
    nested_cl = tup[1][0]
    default = attr.Factory(lambda: [nested_cl()])
    combined_attrs = list(tup[0])
    combined_attrs.append(
        (attr.ib(default=default, type=List[nested_cl]), st.just([nested_cl()]))
    )
    return _create_hyp_class(combined_attrs)


def dict_of_class(tup):
    nested_cl = tup[1][0]
    default = attr.Factory(lambda: {"cls": nested_cl()})
    combined_attrs = list(tup[0])
    combined_attrs.append((attr.ib(default=default), st.just({"cls": nested_cl()})))
    return _create_hyp_class(combined_attrs)


def _create_hyp_nested_strategy(simple_class_strategy):
    """
    Create a recursive attrs class.
    Given a strategy for building (simpler) classes, create and return
    a strategy for building classes that have as an attribute:
        * just the simpler class
        * a list of simpler classes
        * a dict mapping the string "cls" to a simpler class.
    """
    # A strategy producing tuples of the form ([list of attributes], <given
    # class strategy>).
    attrs_and_classes = st.tuples(lists_of_attrs(defaults=True), simple_class_strategy)

    return (
        attrs_and_classes.flatmap(just_class)
        | attrs_and_classes.flatmap(just_class_with_type)
        | attrs_and_classes.flatmap(list_of_class)
        | attrs_and_classes.flatmap(list_of_class_with_type)
        | attrs_and_classes.flatmap(dict_of_class)
        | attrs_and_classes.flatmap(just_frozen_class_with_type)
        | attrs_and_classes.flatmap(just_class_with_type_takes_self)
    )


@st.composite
def bare_attrs(draw, defaults=None, kw_only=None):
    """
    Generate a tuple of an attribute and a strategy that yields values
    appropriate for that attribute.
    """
    default = NOTHING
    if defaults is True or (defaults is None and draw(st.booleans())):
        default = None
    return (
        attr.ib(
            default=default, kw_only=draw(st.booleans()) if kw_only is None else kw_only
        ),
        st.just(None),
    )


@st.composite
def int_attrs(draw, defaults=None, kw_only=None):
    """
    Generate a tuple of an attribute and a strategy that yields ints for that
    attribute.
    """
    default = NOTHING
    if defaults is True or (defaults is None and draw(st.booleans())):
        default = draw(st.integers())
    return (
        attr.ib(
            default=default, kw_only=draw(st.booleans()) if kw_only is None else kw_only
        ),
        st.integers(),
    )


@st.composite
def str_attrs(draw, defaults=None, type_annotations=None, kw_only=None):
    """
    Generate a tuple of an attribute and a strategy that yields strs for that
    attribute.
    """
    default = NOTHING
    if defaults is True or (defaults is None and draw(st.booleans())):
        default = draw(st.text())
    if (type_annotations is None and draw(st.booleans())) or type_annotations:
        type = str
    else:
        type = None
    return (
        attr.ib(
            default=default,
            type=type,
            kw_only=draw(st.booleans()) if kw_only is None else kw_only,
        ),
        st.text(),
    )


@st.composite
def float_attrs(draw, defaults=None, kw_only=None):
    """
    Generate a tuple of an attribute and a strategy that yields floats for that
    attribute.
    """
    default = NOTHING
    if defaults is True or (defaults is None and draw(st.booleans())):
        default = draw(st.floats())
    return (
        attr.ib(
            default=default, kw_only=draw(st.booleans()) if kw_only is None else kw_only
        ),
        st.floats(allow_nan=False),
    )


@st.composite
def dict_attrs(draw, defaults=None, kw_only=None):
    """
    Generate a tuple of an attribute and a strategy that yields dictionaries
    for that attribute. The dictionaries map strings to integers.
    """
    default = NOTHING
    val_strat = st.dictionaries(keys=st.text(), values=st.integers())
    if defaults is True or (defaults is None and draw(st.booleans())):
        default_val = draw(val_strat)
        default = attr.Factory(lambda: default_val)
    return (
        attr.ib(
            default=default, kw_only=draw(st.booleans()) if kw_only is None else kw_only
        ),
        val_strat,
    )


@st.composite
def optional_attrs(draw, defaults=None, kw_only=None):
    """
    Generate a tuple of an attribute and a strategy that yields values
    for that attribute. The strategy generates optional integers.
    """
    default = NOTHING
    val_strat = st.integers() | st.none()
    if defaults is True or (defaults is None and draw(st.booleans())):
        default = draw(val_strat)

    return (
        attr.ib(
            default=default, kw_only=draw(st.booleans()) if kw_only is None else kw_only
        ),
        val_strat,
    )


def simple_attrs(defaults=None, kw_only=None):
    return (
        bare_attrs(defaults, kw_only=kw_only)
        | int_attrs(defaults, kw_only=kw_only)
        | str_attrs(defaults, kw_only=kw_only)
        | float_attrs(defaults, kw_only=kw_only)
        | dict_attrs(defaults, kw_only=kw_only)
        | optional_attrs(defaults, kw_only=kw_only)
    )


def lists_of_attrs(defaults=None, min_size=0, kw_only=None):
    # Python functions support up to 255 arguments.
    return st.lists(
        simple_attrs(defaults, kw_only), min_size=min_size, max_size=10
    ).map(lambda l: sorted(l, key=lambda t: t[0]._default is not NOTHING))


def simple_classes(defaults=None, min_attrs=0, frozen=None, kw_only=None):
    """
    Return a strategy that yields tuples of simple classes and values to
    instantiate them.
    """
    return lists_of_attrs(defaults, min_size=min_attrs, kw_only=kw_only).flatmap(
        lambda attrs_and_strategy: _create_hyp_class(attrs_and_strategy, frozen=frozen)
    )


# Ok, so st.recursive works by taking a base strategy (in this case,
# simple_classes) and a special function. This function receives a strategy,
# and returns another strategy (building on top of the base strategy).
nested_classes = st.recursive(
    simple_classes(defaults=True), _create_hyp_nested_strategy
)



from os import environ

import pytest
from hypothesis import HealthCheck, settings

from cattrs import BaseConverter, Converter


@pytest.fixture(params=(True, False))
def genconverter(request):
    return Converter(detailed_validation=request.param)


@pytest.fixture(params=(True, False))
def converter(request, converter_cls):
    return converter_cls(detailed_validation=request.param)


@pytest.fixture(params=(BaseConverter, Converter), scope="session")
def converter_cls(request):
    return request.param


settings.register_profile(
    "tests", suppress_health_check=(HealthCheck.too_slow,), deadline=None
)
settings.register_profile("fast", settings.get_profile("tests"), max_examples=10)

settings.load_profile("fast" if "FAST" in environ else "tests")



"""Preconfigured converters for bson."""
from base64 import b85decode, b85encode
from datetime import datetime
from typing import Any, Type, TypeVar

from bson import DEFAULT_CODEC_OPTIONS, CodecOptions, ObjectId, decode, encode

from cattrs._compat import AbstractSet, is_mapping
from cattrs.gen import make_mapping_structure_fn

from ..converters import BaseConverter, Converter
from . import validate_datetime

T = TypeVar("T")


class Base85Bytes(bytes):
    """A subclass to help with binary key encoding/decoding."""


class BsonConverter(Converter):
    def dumps(
        self,
        obj: Any,
        unstructure_as: Any = None,
        check_keys: bool = False,
        codec_options: CodecOptions = DEFAULT_CODEC_OPTIONS,
    ) -> bytes:
        return encode(
            self.unstructure(obj, unstructure_as=unstructure_as),
            check_keys=check_keys,
            codec_options=codec_options,
        )

    def loads(
        self,
        data: bytes,
        cl: Type[T],
        codec_options: CodecOptions = DEFAULT_CODEC_OPTIONS,
    ) -> T:
        return self.structure(decode(data, codec_options=codec_options), cl)


def configure_converter(converter: BaseConverter):
    """
    Configure the converter for use with the bson library.

    * sets are serialized as lists
    * byte mapping keys are base85-encoded into strings when unstructuring, and reverse
    * non-string, non-byte mapping keys are coerced into strings when unstructuring
    * a deserialization hook is registered for bson.ObjectId by default
    """

    def gen_unstructure_mapping(cl: Any, unstructure_to=None):
        key_handler = str
        args = getattr(cl, "__args__", None)
        if args:
            if issubclass(args[0], str):
                key_handler = None
            elif issubclass(args[0], bytes):

                def key_handler(k):
                    return b85encode(k).decode("utf8")

        return converter.gen_unstructure_mapping(
            cl, unstructure_to=unstructure_to, key_handler=key_handler
        )

    def gen_structure_mapping(cl: Any):
        args = getattr(cl, "__args__", None)
        if args and issubclass(args[0], bytes):
            h = make_mapping_structure_fn(cl, converter, key_type=Base85Bytes)
        else:
            h = make_mapping_structure_fn(cl, converter)
        return h

    converter.register_structure_hook(Base85Bytes, lambda v, _: b85decode(v))
    converter._unstructure_func.register_func_list(
        [(is_mapping, gen_unstructure_mapping, True)]
    )
    converter._structure_func.register_func_list(
        [(is_mapping, gen_structure_mapping, True)]
    )

    converter.register_structure_hook(datetime, validate_datetime)
    converter.register_structure_hook(ObjectId, lambda v, _: ObjectId(v))


def make_converter(*args: Any, **kwargs: Any) -> BsonConverter:
    kwargs["unstruct_collection_overrides"] = {
        AbstractSet: list,
        **kwargs.get("unstruct_collection_overrides", {}),
    }
    res = BsonConverter(*args, **kwargs)
    configure_converter(res)

    return res



from cattrs._compat import is_py37, is_py38

if is_py37 or is_py38:
    from typing import Dict, List

    List_origin = List
    Dict_origin = Dict


else:
    List_origin = list
    Dict_origin = dict



from collections import deque
from typing import Deque, Dict, Generic, List, Optional, TypeVar, Union

import pytest
from attr import asdict, attrs, define

from cattrs import BaseConverter, Converter
from cattrs._compat import Protocol, is_py39_plus, is_py310_plus, is_py311_plus
from cattrs._generics import deep_copy_with
from cattrs.errors import StructureHandlerNotFoundError
from cattrs.gen._generics import generate_mapping

from ._compat import Dict_origin, List_origin

T = TypeVar("T")
T2 = TypeVar("T2")


def test_deep_copy():
    """Test the deep copying of generic parameters."""
    mapping = {T.__name__: int}
    assert deep_copy_with(Optional[T], mapping) == Optional[int]
    assert (
        deep_copy_with(List_origin[Optional[T]], mapping) == List_origin[Optional[int]]
    )
    mapping = {T.__name__: int, T2.__name__: str}
    assert (
        deep_copy_with(Dict_origin[T2, List_origin[Optional[T]]], mapping)
        == Dict_origin[str, List_origin[Optional[int]]]
    )


@define
class TClass(Generic[T, T2]):
    a: T
    b: T2


@define
class GenericCols(Generic[T]):
    a: T
    b: List[T]
    c: Dict[str, T]


@pytest.mark.parametrize(
    ("t", "t2", "result"),
    (
        (int, str, TClass(1, "a")),
        (str, str, TClass("1", "a")),
        (List[int], str, TClass([1, 2, 3], "a")),
    ),
)
def test_able_to_structure_generics(converter: BaseConverter, t, t2, result):
    res = converter.structure(asdict(result), TClass[t, t2])

    assert res == result


@pytest.mark.parametrize(
    ("t", "result"),
    ((int, GenericCols(1, [2], {"3": 3})), (str, GenericCols("1", ["2"], {"3": "3"}))),
)
@pytest.mark.parametrize("detailed_validation", [True, False])
def test_structure_generics_with_cols(t, result, detailed_validation):
    raw = asdict(result)
    res = Converter(detailed_validation=detailed_validation).structure(
        raw, GenericCols[t]
    )

    assert res == result


@pytest.mark.skipif(not is_py39_plus, reason="3.9+ generics syntax")
@pytest.mark.parametrize(
    ("t", "result"), ((int, (1, [2], {"3": 3})), (str, ("1", ["2"], {"3": "3"})))
)
def test_39_structure_generics_with_cols(t, result):
    @define
    class GenericCols(Generic[T]):
        a: T
        b: list[T]
        c: dict[str, T]

    expected = GenericCols(*result)

    res = Converter().structure(asdict(expected), GenericCols[t])

    assert res == expected


@pytest.mark.parametrize(("t", "result"), ((int, (1, [1, 2, 3])), (int, (1, None))))
def test_structure_nested_generics_with_cols(t, result):
    @define
    class GenericCols(Generic[T]):
        a: T
        b: Optional[List[T]]

    expected = GenericCols(*result)

    res = Converter().structure(asdict(expected), GenericCols[t])

    assert res == expected


@pytest.mark.parametrize(
    ("t", "t2", "result"),
    (
        (TClass[int, int], str, TClass(TClass(1, 2), "a")),
        (List[TClass[int, int]], str, TClass([TClass(1, 2)], "a")),
    ),
)
def test_structure_nested_generics(converter: BaseConverter, t, t2, result):
    res = converter.structure(asdict(result), TClass[t, t2])

    assert res == result


def test_able_to_structure_deeply_nested_generics_gen(converter):
    cl = TClass[TClass[TClass[int, int], int], int]
    result = TClass(TClass(TClass(1, 2), 3), 4)

    res = converter.structure(asdict(result), cl)

    assert res == result


def test_structure_unions_of_generics(converter):
    @attrs(auto_attribs=True)
    class TClass2(Generic[T]):
        c: T

    data = TClass2(c="string")
    res = converter.structure(asdict(data), Union[TClass[int, int], TClass2[str]])
    assert res == data


def test_structure_list_of_generic_unions(converter):
    @attrs(auto_attribs=True)
    class TClass2(Generic[T]):
        c: T

    data = [TClass2(c="string"), TClass(1, 2)]
    res = converter.structure(
        [asdict(x) for x in data], List[Union[TClass[int, int], TClass2[str]]]
    )
    assert res == data


def test_structure_deque_of_generic_unions(converter):
    @attrs(auto_attribs=True)
    class TClass2(Generic[T]):
        c: T

    data = deque((TClass2(c="string"), TClass(1, 2)))
    res = converter.structure(
        [asdict(x) for x in data], Deque[Union[TClass[int, int], TClass2[str]]]
    )
    assert res == data


def test_raises_if_no_generic_params_supplied(
    converter: Union[Converter, BaseConverter]
):
    data = TClass(1, "a")

    with pytest.raises(
        StructureHandlerNotFoundError,
        match="Unsupported type: ~T. Register a structure hook for it.|Missing type for generic argument T, specify it when structuring.",
    ) as exc:
        converter.structure(asdict(data), TClass)

    assert exc.value.type_ is T


def test_unstructure_generic_attrs():
    c = Converter()

    @attrs(auto_attribs=True)
    class Inner(Generic[T]):
        a: T

    @attrs(auto_attribs=True)
    class Outer:
        inner: Inner[int]

    initial = Outer(Inner(1))
    raw = c.unstructure(initial)

    assert raw == {"inner": {"a": 1}}

    new = c.structure(raw, Outer)
    assert initial == new

    @attrs(auto_attribs=True)
    class OuterStr:
        inner: Inner[str]

    assert c.structure(raw, OuterStr) == OuterStr(Inner("1"))


def test_unstructure_deeply_nested_generics():
    c = Converter()

    @define
    class Inner:
        a: int

    @define
    class Outer(Generic[T]):
        inner: T

    initial = Outer[Inner](Inner(1))
    raw = c.unstructure(initial, Outer[Inner])
    assert raw == {"inner": {"a": 1}}

    raw = c.unstructure(initial)
    assert raw == {"inner": {"a": 1}}


def test_unstructure_deeply_nested_generics_list():
    c = Converter()

    @define
    class Inner:
        a: int

    @define
    class Outer(Generic[T]):
        inner: List[T]

    initial = Outer[Inner]([Inner(1)])
    raw = c.unstructure(initial, Outer[Inner])
    assert raw == {"inner": [{"a": 1}]}

    raw = c.unstructure(initial)
    assert raw == {"inner": [{"a": 1}]}


def test_unstructure_protocol():
    c = Converter()

    class Proto(Protocol):
        a: int

    @define
    class Inner:
        a: int

    @define
    class Outer:
        inner: Proto

    initial = Outer(Inner(1))
    raw = c.unstructure(initial, Outer)
    assert raw == {"inner": {"a": 1}}

    raw = c.unstructure(initial)
    assert raw == {"inner": {"a": 1}}


@pytest.mark.skipif(not is_py310_plus, reason="3.10+ union syntax")
def test_roundtrip_generic_with_union() -> None:
    """Generators should handle classes with unions in their names."""
    c = Converter()

    @define
    class A:
        a: int

    @define
    class B:
        b: int

    @define
    class Outer(Generic[T]):
        member: T

    raw = c.unstructure(Outer(A(1)), unstructure_as=Outer[A | B])
    assert c.structure(raw, Outer[A | B]) == Outer((A(1)))


@pytest.mark.skipif(not is_py311_plus, reason="3.11+ only")
def test_generate_typeddict_mapping() -> None:
    from typing import Generic, TypedDict, TypeVar

    T = TypeVar("T")

    class A(TypedDict):
        pass

    assert generate_mapping(A, {}) == {}

    class A(TypedDict, Generic[T]):
        a: T

    assert generate_mapping(A[int], {}) == {T.__name__: int}

    class B(A[int]):
        pass

    assert generate_mapping(B, {}) == {T.__name__: int}



from functools import lru_cache, singledispatch
from typing import Any, Callable, List, Optional, Tuple, Union

import attr


@attr.s
class _DispatchNotFound:
    """A dummy object to help signify a dispatch not found."""

    pass


class MultiStrategyDispatch:
    """
    MultiStrategyDispatch uses a combination of exact-match dispatch,
    singledispatch, and FunctionDispatch.
    """

    __slots__ = (
        "_direct_dispatch",
        "_function_dispatch",
        "_single_dispatch",
        "_generators",
        "_fallback_func",
        "dispatch",
    )

    def __init__(self, fallback_func: Callable[[Any, Any], Any]):
        self._direct_dispatch = {}
        self._function_dispatch = FunctionDispatch()
        self._single_dispatch = singledispatch(_DispatchNotFound)
        self.dispatch = lru_cache(maxsize=None)(self._dispatch)
        self._fallback_func = fallback_func

    def _dispatch(self, typ: Any) -> Callable[[Any, Any], Any]:
        try:
            dispatch = self._single_dispatch.dispatch(typ)
            if dispatch is not _DispatchNotFound:
                return dispatch
        except Exception:
            pass

        direct_dispatch = self._direct_dispatch.get(typ)
        if direct_dispatch is not None:
            return direct_dispatch

        res = self._function_dispatch.dispatch(typ)
        return res if res is not None else self._fallback_func

    def register_cls_list(self, cls_and_handler, direct: bool = False) -> None:
        """Register a class to direct or singledispatch."""
        for cls, handler in cls_and_handler:
            if direct:
                self._direct_dispatch[cls] = handler
            else:
                self._single_dispatch.register(cls, handler)
                self.clear_direct()
        self.dispatch.cache_clear()

    def register_func_list(
        self,
        pred_and_handler: List[
            Union[
                Tuple[Callable[[Any], bool], Any],
                Tuple[Callable[[Any], bool], Any, bool],
            ]
        ],
    ):
        """
        Register a predicate function to determine if the handle
        should be used for the type.
        """
        for tup in pred_and_handler:
            if len(tup) == 2:
                func, handler = tup
                self._function_dispatch.register(func, handler)
            else:
                func, handler, is_gen = tup
                self._function_dispatch.register(func, handler, is_generator=is_gen)
        self.clear_direct()
        self.dispatch.cache_clear()

    def clear_direct(self):
        """Clear the direct dispatch."""
        self._direct_dispatch.clear()

    def clear_cache(self):
        """Clear all caches."""
        self._direct_dispatch.clear()
        self.dispatch.cache_clear()

    def get_num_fns(self) -> int:
        return self._function_dispatch.get_num_fns()

    def copy_to(self, other: "MultiStrategyDispatch", skip: int = 0):
        self._function_dispatch.copy_to(other._function_dispatch, skip=skip)
        for cls, fn in self._single_dispatch.registry.items():
            other._single_dispatch.register(cls, fn)


@attr.s(slots=True)
class FunctionDispatch:
    """
    FunctionDispatch is similar to functools.singledispatch, but
    instead dispatches based on functions that take the type of the
    first argument in the method, and return True or False.

    objects that help determine dispatch should be instantiated objects.
    """

    _handler_pairs: List[
        Tuple[Callable[[Any], bool], Callable[[Any, Any], Any], bool]
    ] = attr.ib(factory=list)

    def register(self, can_handle: Callable[[Any], bool], func, is_generator=False):
        self._handler_pairs.insert(0, (can_handle, func, is_generator))

    def dispatch(self, typ: Any) -> Optional[Callable[[Any, Any], Any]]:
        """
        returns the appropriate handler, for the object passed.
        """
        for can_handle, handler, is_generator in self._handler_pairs:
            # can handle could raise an exception here
            # such as issubclass being called on an instance.
            # it's easier to just ignore that case.
            try:
                ch = can_handle(typ)
            except Exception:
                continue
            if ch:
                if is_generator:
                    return handler(typ)
                else:
                    return handler
        return None

    def get_num_fns(self) -> int:
        return len(self._handler_pairs)

    def copy_to(self, other: "FunctionDispatch", skip: int = 0):
        other._handler_pairs.extend(self._handler_pairs[skip:])



"""Preconfigured converters for pyyaml."""
from cattrs.preconf.pyyaml import configure_converter, make_converter, PyyamlConverter

__all__ = ["configure_converter", "make_converter", "PyyamlConverter"]



"""Line-cache functionality."""
import linecache
import uuid
from typing import Any


def generate_unique_filename(cls: Any, func_name: str, reserve: bool = True) -> str:
    """
    Create a "filename" suitable for a function being generated.
    """
    unique_id = uuid.uuid4()
    extra = ""
    count = 1

    while True:
        unique_filename = "<cattrs generated {0} {1}.{2}{3}>".format(
            func_name, cls.__module__, getattr(cls, "__qualname__", cls.__name__), extra
        )
        if not reserve:
            return unique_filename
        # To handle concurrency we essentially "reserve" our spot in
        # the linecache with a dummy line.  The caller can then
        # set this value correctly.
        cache_line = (1, None, (str(unique_id),), unique_filename)
        if linecache.cache.setdefault(unique_filename, cache_line) == cache_line:
            return unique_filename

        # Looks like this spot is taken. Try again.
        count += 1
        extra = "-{0}".format(count)



from __future__ import annotations

import linecache
import re
from dataclasses import is_dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    Mapping,
    Optional,
    Tuple,
    Type,
    TypeVar,
)

import attr
from attr import NOTHING, resolve_types

from .._compat import (
    _adapted_fields,
    get_args,
    get_origin,
    is_annotated,
    is_bare,
    is_bare_final,
    is_generic,
)
from .._generics import deep_copy_with
from ..errors import (
    AttributeValidationNote,
    ClassValidationError,
    ForbiddenExtraKeysError,
    IterableValidationError,
    IterableValidationNote,
    StructureHandlerNotFoundError,
)
from ._consts import AttributeOverride, already_generating, neutral
from ._generics import generate_mapping
from ._lc import generate_unique_filename
from ._shared import find_structure_handler

if TYPE_CHECKING:  # pragma: no cover
    from cattr.converters import BaseConverter


def override(
    omit_if_default: Optional[bool] = None,
    rename: Optional[str] = None,
    omit: bool = False,
    struct_hook: Optional[Callable[[Any, Any], Any]] = None,
    unstruct_hook: Optional[Callable[[Any], Any]] = None,
):
    return AttributeOverride(omit_if_default, rename, omit, struct_hook, unstruct_hook)


T = TypeVar("T")


def make_dict_unstructure_fn(
    cl: Type[T],
    converter: "BaseConverter",
    _cattrs_omit_if_default: bool = False,
    _cattrs_use_linecache: bool = True,
    **kwargs: AttributeOverride,
) -> Callable[[T], Dict[str, Any]]:
    """
    Generate a specialized dict unstructuring function for an attrs class or a
    dataclass.
    """
    origin = get_origin(cl)
    attrs = _adapted_fields(origin or cl)  # type: ignore

    if any(isinstance(a.type, str) for a in attrs):
        # PEP 563 annotations - need to be resolved.
        resolve_types(cl)

    mapping = {}
    if is_generic(cl):
        mapping = generate_mapping(cl, mapping)

        for base in getattr(origin, "__orig_bases__", ()):
            if is_generic(base) and not str(base).startswith("typing.Generic"):
                mapping = generate_mapping(base, mapping)
                break
        if origin is not None:
            cl = origin

    cl_name = cl.__name__
    fn_name = "unstructure_" + cl_name
    globs = {}
    lines = []
    invocation_lines = []
    internal_arg_parts = {}

    # We keep track of what we're generating to help with recursive
    # class graphs.
    try:
        working_set = already_generating.working_set
    except AttributeError:
        working_set = set()
        already_generating.working_set = working_set
    if cl in working_set:
        raise RecursionError()
    else:
        working_set.add(cl)

    try:
        for a in attrs:
            attr_name = a.name
            override = kwargs.pop(attr_name, neutral)
            if override.omit:
                continue
            kn = attr_name if override.rename is None else override.rename
            d = a.default

            # For each attribute, we try resolving the type here and now.
            # If a type is manually overwritten, this function should be
            # regenerated.
            handler = None
            if override.unstruct_hook is not None:
                handler = override.unstruct_hook
            else:
                if a.type is not None:
                    t = a.type
                    if isinstance(t, TypeVar):
                        if t.__name__ in mapping:
                            t = mapping[t.__name__]
                        else:
                            handler = converter.unstructure
                    elif is_generic(t) and not is_bare(t) and not is_annotated(t):
                        t = deep_copy_with(t, mapping)

                    if handler is None:
                        if (
                            is_bare_final(t)
                            and a.default is not NOTHING
                            and not isinstance(a.default, attr.Factory)
                        ):
                            # This is a special case where we can use the
                            # type of the default to dispatch on.
                            t = a.default.__class__
                        try:
                            handler = converter._unstructure_func.dispatch(t)
                        except RecursionError:
                            # There's a circular reference somewhere down the line
                            handler = converter.unstructure
                else:
                    handler = converter.unstructure

            is_identity = handler == converter._unstructure_identity

            if not is_identity:
                unstruct_handler_name = f"__c_unstr_{attr_name}"
                globs[unstruct_handler_name] = handler
                internal_arg_parts[unstruct_handler_name] = handler
                invoke = f"{unstruct_handler_name}(instance.{attr_name})"
            else:
                invoke = f"instance.{attr_name}"

            if d is not attr.NOTHING and (
                (_cattrs_omit_if_default and override.omit_if_default is not False)
                or override.omit_if_default
            ):
                def_name = f"__c_def_{attr_name}"

                if isinstance(d, attr.Factory):
                    globs[def_name] = d.factory
                    internal_arg_parts[def_name] = d.factory
                    if d.takes_self:
                        lines.append(
                            f"  if instance.{attr_name} != {def_name}(instance):"
                        )
                    else:
                        lines.append(f"  if instance.{attr_name} != {def_name}():")
                    lines.append(f"    res['{kn}'] = {invoke}")
                else:
                    globs[def_name] = d
                    internal_arg_parts[def_name] = d
                    lines.append(f"  if instance.{attr_name} != {def_name}:")
                    lines.append(f"    res['{kn}'] = {invoke}")

            else:
                # No default or no override.
                invocation_lines.append(f"'{kn}': {invoke},")

        internal_arg_line = ", ".join([f"{i}={i}" for i in internal_arg_parts])
        if internal_arg_line:
            internal_arg_line = f", {internal_arg_line}"
        for k, v in internal_arg_parts.items():
            globs[k] = v

        total_lines = (
            [f"def {fn_name}(instance{internal_arg_line}):"]
            + ["  res = {"]
            + [f"    {line}" for line in invocation_lines]
            + ["  }"]
            + lines
            + ["  return res"]
        )
        script = "\n".join(total_lines)

        fname = generate_unique_filename(
            cl, "unstructure", reserve=_cattrs_use_linecache
        )

        eval(compile(script, fname, "exec"), globs)

        fn = globs[fn_name]
        if _cattrs_use_linecache:
            linecache.cache[fname] = len(script), None, total_lines, fname
    finally:
        working_set.remove(cl)

    return fn


DictStructureFn = Callable[[Mapping[str, Any], Any], T]


def make_dict_structure_fn(
    cl: Type[T],
    converter: "BaseConverter",
    _cattrs_forbid_extra_keys: bool = False,
    _cattrs_use_linecache: bool = True,
    _cattrs_prefer_attrib_converters: bool = False,
    _cattrs_detailed_validation: bool = True,
    **kwargs: AttributeOverride,
) -> DictStructureFn[T]:
    """Generate a specialized dict structuring function for an attrs class."""

    mapping = {}
    if is_generic(cl):
        base = get_origin(cl)
        mapping = generate_mapping(cl, mapping)
        if base is not None:
            cl = base

    for base in getattr(cl, "__orig_bases__", ()):
        if is_generic(base) and not str(base).startswith("typing.Generic"):
            mapping = generate_mapping(base, mapping)
            break

    if isinstance(cl, TypeVar):
        cl = mapping.get(cl.__name__, cl)

    cl_name = cl.__name__
    fn_name = "structure_" + cl_name

    # We have generic parameters and need to generate a unique name for the function
    for p in getattr(cl, "__parameters__", ()):
        # This is nasty, I am not sure how best to handle `typing.List[str]` or `TClass[int, int]` as a parameter type here
        try:
            name_base = mapping[p.__name__]
        except KeyError:
            raise StructureHandlerNotFoundError(
                f"Missing type for generic argument {p.__name__}, specify it when structuring.",
                p,
            ) from None
        name = getattr(name_base, "__name__", None) or str(name_base)
        # `<>` can be present in lambdas
        # `|` can be present in unions
        name = re.sub(r"[\[\.\] ,<>]", "_", name)
        name = re.sub(r"\|", "u", name)
        fn_name += f"_{name}"

    internal_arg_parts = {"__cl": cl}
    globs = {}
    lines = []
    post_lines = []
    invocation_lines = []

    attrs = _adapted_fields(cl)
    is_dc = is_dataclass(cl)

    if any(isinstance(a.type, str) for a in attrs):
        # PEP 563 annotations - need to be resolved.
        resolve_types(cl)

    allowed_fields = set()
    if _cattrs_forbid_extra_keys:
        globs["__c_a"] = allowed_fields
        globs["__c_feke"] = ForbiddenExtraKeysError

    if _cattrs_detailed_validation:
        lines.append("  res = {}")
        lines.append("  errors = []")
        invocation_lines.append("**res,")
        internal_arg_parts["__c_cve"] = ClassValidationError
        internal_arg_parts["__c_avn"] = AttributeValidationNote
        for a in attrs:
            an = a.name
            override = kwargs.get(an, neutral)
            if override.omit:
                continue
            t = a.type
            if isinstance(t, TypeVar):
                t = mapping.get(t.__name__, t)
            elif is_generic(t) and not is_bare(t) and not is_annotated(t):
                t = deep_copy_with(t, mapping)

            # For each attribute, we try resolving the type here and now.
            # If a type is manually overwritten, this function should be
            # regenerated.
            if override.struct_hook is not None:
                # If the user has requested an override, just use that.
                handler = override.struct_hook
            else:
                handler = find_structure_handler(
                    a, t, converter, _cattrs_prefer_attrib_converters
                )

            struct_handler_name = f"__c_structure_{an}"
            internal_arg_parts[struct_handler_name] = handler

            ian = an if (is_dc or an[0] != "_") else an[1:]
            kn = an if override.rename is None else override.rename
            allowed_fields.add(kn)
            i = "  "
            if a.default is not NOTHING:
                lines.append(f"{i}if '{kn}' in o:")
                i = f"{i}  "
            lines.append(f"{i}try:")
            i = f"{i}  "
            type_name = f"__c_type_{an}"
            internal_arg_parts[type_name] = t
            if handler:
                if handler == converter._structure_call:
                    internal_arg_parts[struct_handler_name] = t
                    lines.append(f"{i}res['{ian}'] = {struct_handler_name}(o['{kn}'])")
                else:
                    type_name = f"__c_type_{an}"
                    internal_arg_parts[type_name] = t
                    lines.append(
                        f"{i}res['{ian}'] = {struct_handler_name}(o['{kn}'], {type_name})"
                    )
            else:
                lines.append(f"{i}res['{ian}'] = o['{kn}']")
            i = i[:-2]
            lines.append(f"{i}except Exception as e:")
            i = f"{i}  "
            lines.append(
                f'{i}e.__notes__ = getattr(e, \'__notes__\', []) + [__c_avn("Structuring class {cl.__qualname__} @ attribute {an}", "{an}", __c_type_{an})]'
            )
            lines.append(f"{i}errors.append(e)")

        if _cattrs_forbid_extra_keys:
            post_lines += [
                "  unknown_fields = set(o.keys()) - __c_a",
                "  if unknown_fields:",
                "    errors.append(__c_feke('', __cl, unknown_fields))",
            ]

        post_lines.append(
            f"  if errors: raise __c_cve('While structuring ' + {cl.__name__!r}, errors, __cl)"
        )
        instantiation_lines = (
            ["  try:"]
            + ["    return __cl("]
            + [f"      {line}" for line in invocation_lines]
            + ["    )"]
            + [
                f"  except Exception as exc: raise __c_cve('While structuring ' + {cl.__name__!r}, [exc], __cl)"
            ]
        )
    else:
        non_required = []
        # The first loop deals with required args.
        for a in attrs:
            an = a.name
            override = kwargs.get(an, neutral)
            if override.omit:
                continue
            if a.default is not NOTHING:
                non_required.append(a)
                continue
            t = a.type
            if isinstance(t, TypeVar):
                t = mapping.get(t.__name__, t)
            elif is_generic(t) and not is_bare(t) and not is_annotated(t):
                t = deep_copy_with(t, mapping)

            # For each attribute, we try resolving the type here and now.
            # If a type is manually overwritten, this function should be
            # regenerated.
            if override.struct_hook is not None:
                # If the user has requested an override, just use that.
                handler = override.struct_hook
            else:
                handler = find_structure_handler(
                    a, t, converter, _cattrs_prefer_attrib_converters
                )

            kn = an if override.rename is None else override.rename
            allowed_fields.add(kn)

            if handler:
                struct_handler_name = f"__c_structure_{an}"
                internal_arg_parts[struct_handler_name] = handler
                if handler == converter._structure_call:
                    internal_arg_parts[struct_handler_name] = t
                    invocation_line = f"{struct_handler_name}(o['{kn}']),"
                else:
                    type_name = f"__c_type_{an}"
                    internal_arg_parts[type_name] = t
                    invocation_line = f"{struct_handler_name}(o['{kn}'], {type_name}),"
            else:
                invocation_line = f"o['{kn}'],"

            if a.kw_only:
                ian = an if (is_dc or an[0] != "_") else an[1:]
                invocation_line = f"{ian}={invocation_line}"
            invocation_lines.append(invocation_line)

        # The second loop is for optional args.
        if non_required:
            invocation_lines.append("**res,")
            lines.append("  res = {}")

            for a in non_required:
                an = a.name
                override = kwargs.get(an, neutral)
                t = a.type
                if isinstance(t, TypeVar):
                    t = mapping.get(t.__name__, t)
                elif is_generic(t) and not is_bare(t) and not is_annotated(t):
                    t = deep_copy_with(t, mapping)

                # For each attribute, we try resolving the type here and now.
                # If a type is manually overwritten, this function should be
                # regenerated.
                if override.struct_hook is not None:
                    # If the user has requested an override, just use that.
                    handler = override.struct_hook
                else:
                    handler = find_structure_handler(
                        a, t, converter, _cattrs_prefer_attrib_converters
                    )

                struct_handler_name = f"__c_structure_{an}"
                internal_arg_parts[struct_handler_name] = handler

                ian = an if (is_dc or an[0] != "_") else an[1:]
                kn = an if override.rename is None else override.rename
                allowed_fields.add(kn)
                post_lines.append(f"  if '{kn}' in o:")
                if handler:
                    if handler == converter._structure_call:
                        internal_arg_parts[struct_handler_name] = t
                        post_lines.append(
                            f"    res['{ian}'] = {struct_handler_name}(o['{kn}'])"
                        )
                    else:
                        type_name = f"__c_type_{an}"
                        internal_arg_parts[type_name] = t
                        post_lines.append(
                            f"    res['{ian}'] = {struct_handler_name}(o['{kn}'], {type_name})"
                        )
                else:
                    post_lines.append(f"    res['{ian}'] = o['{kn}']")
        instantiation_lines = (
            ["  return __cl("] + [f"    {line}" for line in invocation_lines] + ["  )"]
        )

        if _cattrs_forbid_extra_keys:
            post_lines += [
                "  unknown_fields = set(o.keys()) - __c_a",
                "  if unknown_fields:",
                "    raise __c_feke('', __cl, unknown_fields)",
            ]

    # At the end, we create the function header.
    internal_arg_line = ", ".join([f"{i}={i}" for i in internal_arg_parts])
    for k, v in internal_arg_parts.items():
        globs[k] = v

    total_lines = (
        [f"def {fn_name}(o, _, *, {internal_arg_line}):"]
        + lines
        + post_lines
        + instantiation_lines
    )

    fname = generate_unique_filename(cl, "structure", reserve=_cattrs_use_linecache)
    script = "\n".join(total_lines)
    eval(compile(script, fname, "exec"), globs)
    if _cattrs_use_linecache:
        linecache.cache[fname] = len(script), None, total_lines, fname

    return globs[fn_name]


IterableUnstructureFn = Callable[[Iterable[Any]], Any]


def make_iterable_unstructure_fn(
    cl: Any, converter: "BaseConverter", unstructure_to: Any = None
) -> IterableUnstructureFn:
    """Generate a specialized unstructure function for an iterable."""
    handler = converter.unstructure

    fn_name = "unstructure_iterable"

    # Let's try fishing out the type args
    # Unspecified tuples have `__args__` as empty tuples, so guard
    # against IndexError.
    if getattr(cl, "__args__", None) not in (None, ()):
        type_arg = cl.__args__[0]
        # We don't know how to handle the TypeVar on this level,
        # so we skip doing the dispatch here.
        if not isinstance(type_arg, TypeVar):
            handler = converter._unstructure_func.dispatch(type_arg)

    globs = {"__cattr_seq_cl": unstructure_to or cl, "__cattr_u": handler}
    lines = []

    lines.append(f"def {fn_name}(iterable):")
    lines.append("    res = __cattr_seq_cl(__cattr_u(i) for i in iterable)")

    total_lines = lines + ["    return res"]

    eval(compile("\n".join(total_lines), "", "exec"), globs)

    fn = globs[fn_name]

    return fn


HeteroTupleUnstructureFn = Callable[[Tuple[Any, ...]], Any]


def make_hetero_tuple_unstructure_fn(
    cl: Any, converter: "BaseConverter", unstructure_to: Any = None
) -> HeteroTupleUnstructureFn:
    """Generate a specialized unstructure function for a heterogenous tuple."""
    fn_name = "unstructure_tuple"

    type_args = get_args(cl)

    # We can do the dispatch here and now.
    handlers = [
        converter._unstructure_func.dispatch(type_arg) for type_arg in type_args
    ]

    globs = {f"__cattr_u_{i}": h for i, h in enumerate(handlers)}
    if unstructure_to is not tuple:
        globs["__cattr_seq_cl"] = unstructure_to or cl
    lines = []

    lines.append(f"def {fn_name}(tup):")
    if unstructure_to is not tuple:
        lines.append("    res = __cattr_seq_cl((")
    else:
        lines.append("    res = (")
    for i in range(len(handlers)):
        if handlers[i] == converter._unstructure_identity:
            lines.append(f"        tup[{i}],")
        else:
            lines.append(f"        __cattr_u_{i}(tup[{i}]),")

    if unstructure_to is not tuple:
        lines.append("    ))")
    else:
        lines.append("    )")

    total_lines = lines + ["    return res"]

    eval(compile("\n".join(total_lines), "", "exec"), globs)

    fn = globs[fn_name]

    return fn


MappingUnstructureFn = Callable[[Mapping[Any, Any]], Any]


def make_mapping_unstructure_fn(
    cl: Any,
    converter: "BaseConverter",
    unstructure_to: Any = None,
    key_handler: Optional[Callable[[Any, Optional[Any]], Any]] = None,
) -> MappingUnstructureFn:
    """Generate a specialized unstructure function for a mapping."""
    kh = key_handler or converter.unstructure
    val_handler = converter.unstructure

    fn_name = "unstructure_mapping"

    # Let's try fishing out the type args.
    if getattr(cl, "__args__", None) is not None:
        args = get_args(cl)
        if len(args) == 2:
            key_arg, val_arg = args
        else:
            # Probably a Counter
            key_arg, val_arg = args, Any
        # We can do the dispatch here and now.
        kh = key_handler or converter._unstructure_func.dispatch(key_arg)
        if kh == converter._unstructure_identity:
            kh = None

        val_handler = converter._unstructure_func.dispatch(val_arg)
        if val_handler == converter._unstructure_identity:
            val_handler = None

    globs = {
        "__cattr_mapping_cl": unstructure_to or cl,
        "__cattr_k_u": kh,
        "__cattr_v_u": val_handler,
    }

    k_u = "__cattr_k_u(k)" if kh is not None else "k"
    v_u = "__cattr_v_u(v)" if val_handler is not None else "v"

    lines = []

    lines.append(f"def {fn_name}(mapping):")
    lines.append(
        f"    res = __cattr_mapping_cl(({k_u}, {v_u}) for k, v in mapping.items())"
    )

    total_lines = lines + ["    return res"]

    eval(compile("\n".join(total_lines), "", "exec"), globs)

    fn = globs[fn_name]

    return fn


MappingStructureFn = Callable[[Mapping[Any, Any], Any], T]


def make_mapping_structure_fn(
    cl: Type[T],
    converter: "BaseConverter",
    structure_to: Type = dict,
    key_type=NOTHING,
    val_type=NOTHING,
    detailed_validation: bool = True,
) -> MappingStructureFn[T]:
    """Generate a specialized unstructure function for a mapping."""
    fn_name = "structure_mapping"

    globs: Dict[str, Type] = {"__cattr_mapping_cl": structure_to}

    lines = []
    internal_arg_parts = {}

    # Let's try fishing out the type args.
    if not is_bare(cl):
        args = get_args(cl)
        if len(args) == 2:
            key_arg_cand, val_arg_cand = args
            if key_type is NOTHING:
                key_type = key_arg_cand
            if val_type is NOTHING:
                val_type = val_arg_cand
        else:
            if key_type is not NOTHING and val_type is NOTHING:
                (val_type,) = args
            elif key_type is NOTHING and val_type is not NOTHING:
                (key_type,) = args
            else:
                # Probably a Counter
                (key_type,) = args
                val_type = Any

        is_bare_dict = val_type is Any and key_type is Any
        if not is_bare_dict:
            # We can do the dispatch here and now.
            key_handler = converter._structure_func.dispatch(key_type)
            if key_handler == converter._structure_call:
                key_handler = key_type

            val_handler = converter._structure_func.dispatch(val_type)
            if val_handler == converter._structure_call:
                val_handler = val_type

            globs["__cattr_k_t"] = key_type
            globs["__cattr_v_t"] = val_type
            globs["__cattr_k_s"] = key_handler
            globs["__cattr_v_s"] = val_handler
            k_s = (
                "__cattr_k_s(k, __cattr_k_t)"
                if key_handler != key_type
                else "__cattr_k_s(k)"
            )
            v_s = (
                "__cattr_v_s(v, __cattr_v_t)"
                if val_handler != val_type
                else "__cattr_v_s(v)"
            )
    else:
        is_bare_dict = True

    if is_bare_dict:
        # No args, it's a bare dict.
        lines.append("  res = dict(mapping)")
    else:
        if detailed_validation:
            internal_arg_parts["IterableValidationError"] = IterableValidationError
            internal_arg_parts["IterableValidationNote"] = IterableValidationNote
            internal_arg_parts["val_type"] = (
                val_type if val_type is not NOTHING else Any
            )
            internal_arg_parts["key_type"] = (
                key_type if key_type is not NOTHING else Any
            )
            globs["enumerate"] = enumerate

            lines.append("  res = {}; errors = []")
            lines.append("  for ix, (k, v) in enumerate(mapping.items()):")
            lines.append("    try:")
            lines.append(f"      value = {v_s}")
            lines.append("    except Exception as e:")
            lines.append(
                "      e.__notes__ = getattr(e, '__notes__', []) + [IterableValidationNote('Structuring mapping value @ key ' + repr(k), k, val_type)]"
            )
            lines.append("      errors.append(e)")
            lines.append("      continue")
            lines.append("    try:")
            lines.append(f"      key = {k_s}")
            lines.append("      res[key] = value")
            lines.append("    except Exception as e:")
            lines.append(
                "      e.__notes__ = getattr(e, '__notes__', []) + [IterableValidationNote('Structuring mapping key @ key ' + repr(k), k, key_type)]"
            )
            lines.append("      errors.append(e)")
            lines.append("  if errors:")
            lines.append(
                f"    raise IterableValidationError('While structuring ' + {repr(cl)!r}, errors, __cattr_mapping_cl)"
            )
        else:
            lines.append(f"  res = {{{k_s}: {v_s} for k, v in mapping.items()}}")
    if structure_to is not dict:
        lines.append("  res = __cattr_mapping_cl(res)")

    internal_arg_line = ", ".join([f"{i}={i}" for i in internal_arg_parts])
    if internal_arg_line:
        internal_arg_line = f", {internal_arg_line}"
    for k, v in internal_arg_parts.items():
        globs[k] = v

    def_line = f"def {fn_name}(mapping, _{internal_arg_line}):"
    total_lines = [def_line] + lines + ["  return res"]
    script = "\n".join(total_lines)

    eval(compile(script, "", "exec"), globs)

    fn = globs[fn_name]

    return fn



"""Strategies for typed dicts."""
from datetime import datetime
from string import ascii_lowercase
from typing import Any, Dict, Generic, List, Optional, Set, Tuple, TypeVar

from attr import NOTHING
from hypothesis.strategies import (
    DrawFn,
    SearchStrategy,
    booleans,
    composite,
    datetimes,
    integers,
    just,
    lists,
    sets,
    text,
)

from cattrs._compat import ExtensionsTypedDict, NotRequired, Required, TypedDict

from .untyped import gen_attr_names

# Type aliases for readability
TypedDictType = type
T1 = TypeVar("T1")
T2 = TypeVar("T2")
T3 = TypeVar("T3")


@composite
def int_attributes(
    draw: DrawFn, total: bool = True, not_required: bool = False
) -> Tuple[int, SearchStrategy, SearchStrategy]:
    if total:
        if not_required and draw(booleans()):
            return (NotRequired[int], integers() | just(NOTHING), text(ascii_lowercase))
        else:
            return int, integers(), text(ascii_lowercase)
    else:
        if not_required and draw(booleans()):
            return Required[int], integers(), text(ascii_lowercase)
        else:
            return int, integers() | just(NOTHING), text(ascii_lowercase)


@composite
def datetime_attributes(
    draw: DrawFn, total: bool = True, not_required: bool = False
) -> Tuple[datetime, SearchStrategy, SearchStrategy]:
    success_strat = datetimes().map(lambda dt: dt.replace(microsecond=0))
    type = datetime
    strat = success_strat if total else success_strat | just(NOTHING)
    if not_required and draw(booleans()):
        if total:
            type = NotRequired[type]
            strat = success_strat | just(NOTHING)
        else:
            type = Required[type]
            strat = success_strat
    return (type, strat, text(ascii_lowercase))


@composite
def list_of_int_attributes(
    draw: DrawFn, total: bool = True, not_required: bool = False
) -> Tuple[List[int], SearchStrategy, SearchStrategy]:
    if total:
        if not_required and draw(booleans()):
            return (
                NotRequired[List[int]],
                lists(integers()) | just(NOTHING),
                text(ascii_lowercase).map(lambda v: [v]),
            )
        else:
            return (
                List[int],
                lists(integers()),
                text(ascii_lowercase).map(lambda v: [v]),
            )
    else:
        if not_required and draw(booleans()):
            return (
                Required[List[int]],
                lists(integers()),
                text(ascii_lowercase).map(lambda v: [v]),
            )
        else:
            return (
                List[int],
                lists(integers()) | just(NOTHING),
                text(ascii_lowercase).map(lambda v: [v]),
            )


@composite
def simple_typeddicts(
    draw: DrawFn,
    total: Optional[bool] = None,
    not_required: bool = False,
    min_attrs: int = 0,
    typeddict_cls: Optional[Any] = None,
) -> Tuple[TypedDictType, dict]:
    """Generate simple typed dicts.

    :param total: Generate the given totality dicts (default = random)
    """
    if total is None:
        total = draw(booleans())

    attrs = draw(
        lists(
            int_attributes(total, not_required)
            | list_of_int_attributes(total, not_required)
            | datetime_attributes(total, not_required),
            min_size=min_attrs,
        )
    )

    attrs_dict = {n: attr[0] for n, attr in zip(gen_attr_names(), attrs)}
    success_payload = {}
    for n, a in zip(attrs_dict, attrs):
        v = draw(a[1])
        if v is not NOTHING:
            success_payload[n] = v

    if typeddict_cls is None:
        cls = (TypedDict if draw(booleans()) else ExtensionsTypedDict)(
            "HypTypedDict", attrs_dict, total=total
        )
    else:
        cls = typeddict_cls

    if draw(booleans()):

        class InheritedTypedDict(cls):
            inherited: int

        cls = InheritedTypedDict
        success_payload["inherited"] = draw(integers())

    return (cls, success_payload)


@composite
def simple_typeddicts_with_extra_keys(
    draw: DrawFn, total: Optional[bool] = None, typeddict_cls: Optional[Any] = None
) -> Tuple[TypedDictType, dict, Set[str]]:
    """Generate TypedDicts, with the instances having extra keys."""
    cls, success = draw(simple_typeddicts(total, typeddict_cls=typeddict_cls))

    # The normal attributes are 2 characters or less.
    extra_keys = draw(sets(text(ascii_lowercase, min_size=3, max_size=3)))
    success.update({k: 1 for k in extra_keys})

    return cls, success, extra_keys


@composite
def generic_typeddicts(
    draw: DrawFn, total: Optional[bool] = None
) -> Tuple[TypedDictType, dict]:
    """Generate generic typed dicts.

    :param total: Generate the given totality dicts (default = random)
    """
    if total is None:
        total = draw(booleans())

    attrs = draw(
        lists(
            int_attributes(total)
            | list_of_int_attributes(total)
            | datetime_attributes(total),
            min_size=1,
        )
    )

    attrs_dict = {n: attr[0] for n, attr in zip(gen_attr_names(), attrs)}
    success_payload = {}
    for n, a in zip(attrs_dict, attrs):
        v = draw(a[1])
        if v is not NOTHING:
            success_payload[n] = v

    # We choose up to 3 attributes and make them generic.
    generic_attrs = draw(
        lists(integers(0, len(attrs) - 1), min_size=1, max_size=3, unique=True)
    )
    generics = []
    actual_types = []
    for ix, (attr_name, attr_type) in enumerate(list(attrs_dict.items())):
        if ix in generic_attrs:
            typevar = TypeVar(f"T{ix+1}")
            generics.append(typevar)
            actual_types.append(attr_type)
            attrs_dict[attr_name] = typevar

    cls = make_typeddict(
        "HypTypedDict", attrs_dict, total=total, bases=[Generic[tuple(generics)]]
    )

    if draw(booleans()):

        class InheritedTypedDict(cls[tuple(actual_types)]):
            inherited: int

        cls = InheritedTypedDict
        success_payload["inherited"] = draw(integers())
    else:
        cls = cls[tuple(actual_types)]

    return (cls, success_payload)


def make_typeddict(
    cls_name: str, attrs: Dict[str, type], total: bool = True, bases: List = []
) -> TypedDictType:
    globs = {"TypedDict": TypedDict}
    lines = []

    bases_snippet = ",".join(f"_base{ix}" for ix in range(len(bases)))
    for ix, base in enumerate(bases):
        globs[f"_base{ix}"] = base
    if bases_snippet:
        bases_snippet = f", {bases_snippet}"

    lines.append(f"class {cls_name}(TypedDict{bases_snippet},total={total}):")
    for n, t in attrs.items():
        # Strip the initial underscore if present, to prevent mangling.
        trimmed = n[1:] if n.startswith("_") else n
        globs[f"_{trimmed}_type"] = t
        lines.append(f"  {n}: _{trimmed}_type")

    script = "\n".join(lines)
    eval(compile(script, "name", "exec"), globs)

    cls = globs[cls_name]

    return cls



from __future__ import annotations

from typing import Dict, Type, TypeVar

from .._compat import get_args, get_origin, is_generic


def generate_mapping(cl: Type, old_mapping: Dict[str, type] = {}) -> Dict[str, type]:
    mapping = {}

    origin = get_origin(cl)

    if origin is not None:
        # To handle the cases where classes in the typing module are using
        # the GenericAlias structure but aren’t a Generic and hence
        # end up in this function but do not have an `__parameters__`
        # attribute. These classes are interface types, for example
        # `typing.Hashable`.
        parameters = getattr(get_origin(cl), "__parameters__", None)
        if parameters is None:
            return dict(old_mapping)

        for p, t in zip(parameters, get_args(cl)):
            if isinstance(t, TypeVar):
                continue
            mapping[p.__name__] = t

        if not mapping:
            return dict(old_mapping)
    elif is_generic(cl):
        # Origin is None, so this may be a subclass of a generic class.
        orig_bases = cl.__orig_bases__
        for base in orig_bases:
            if not hasattr(base, "__args__"):
                continue
            base_args = base.__args__
            if not hasattr(base.__origin__, "__parameters__"):
                continue
            base_params = base.__origin__.__parameters__
            for param, arg in zip(base_params, base_args):
                mapping[param.__name__] = arg

    return mapping



"""Preconfigured converters for msgpack."""
from datetime import datetime, timezone
from typing import Any, Type, TypeVar

from msgpack import dumps, loads

from cattrs._compat import AbstractSet

from ..converters import BaseConverter, Converter

T = TypeVar("T")


class MsgpackConverter(Converter):
    def dumps(self, obj: Any, unstructure_as: Any = None, **kwargs: Any) -> bytes:
        return dumps(self.unstructure(obj, unstructure_as=unstructure_as), **kwargs)

    def loads(self, data: bytes, cl: Type[T], **kwargs: Any) -> T:
        return self.structure(loads(data, **kwargs), cl)


def configure_converter(converter: BaseConverter):
    """
    Configure the converter for use with the msgpack library.

    * datetimes are serialized as timestamp floats
    * sets are serialized as lists
    """
    converter.register_unstructure_hook(datetime, lambda v: v.timestamp())
    converter.register_structure_hook(
        datetime, lambda v, _: datetime.fromtimestamp(v, timezone.utc)
    )


def make_converter(*args: Any, **kwargs: Any) -> MsgpackConverter:
    kwargs["unstruct_collection_overrides"] = {
        AbstractSet: list,
        **kwargs.get("unstruct_collection_overrides", {}),
    }
    res = MsgpackConverter(*args, **kwargs)
    configure_converter(res)

    return res



"""Tests for functionality from the gen module."""
import linecache
from traceback import format_exc

from attr import define

from cattrs import Converter
from cattrs.gen import make_dict_structure_fn, make_dict_unstructure_fn


def test_structure_linecache():
    """Linecaching for structuring should work."""

    @define
    class A:
        a: int

    c = Converter(detailed_validation=False)
    try:
        c.structure({"a": "test"}, A)
    except ValueError:
        res = format_exc()
        assert "'a'" in res


def test_unstructure_linecache():
    """Linecaching for unstructuring should work."""

    @define
    class Inner:
        a: int

    @define
    class Outer:
        inner: Inner

    c = Converter()
    try:
        c.unstructure(Outer({}))
    except AttributeError:
        res = format_exc()
        assert "'a'" in res


def test_no_linecache():
    """Linecaching should be disableable."""

    @define
    class A:
        a: int

    c = Converter()
    before = len(linecache.cache)
    c.structure(c.unstructure(A(1)), A)
    after = len(linecache.cache)

    assert after == before + 2

    @define
    class B:
        a: int

    before = len(linecache.cache)
    c.register_structure_hook(
        B, make_dict_structure_fn(B, c, _cattrs_use_linecache=False)
    )
    c.register_unstructure_hook(
        B, make_dict_unstructure_fn(B, c, _cattrs_use_linecache=False)
    )
    c.structure(c.unstructure(B(1)), B)

    assert len(linecache.cache) == before



from cattrs.errors import (
    BaseValidationError,
    ClassValidationError,
    ForbiddenExtraKeysError,
    IterableValidationError,
    StructureHandlerNotFoundError,
)

__all__ = [
    "BaseValidationError",
    "ClassValidationError",
    "ForbiddenExtraKeysError",
    "IterableValidationError",
    "StructureHandlerNotFoundError",
]



"""Tests for NewTypes."""
from typing import NewType

import pytest

from cattrs import BaseConverter

PositiveIntNewType = NewType("PositiveIntNewType", int)
BigPositiveIntNewType = NewType("BigPositiveIntNewType", PositiveIntNewType)


def test_newtype_structure_hooks(genconverter: BaseConverter):
    """NewTypes should work with `register_structure_hook`."""

    assert genconverter.structure("0", int) == 0
    assert genconverter.structure("0", PositiveIntNewType) == 0
    assert genconverter.structure("0", BigPositiveIntNewType) == 0

    genconverter.register_structure_hook(
        PositiveIntNewType, lambda v, _: int(v) if int(v) > 0 else 1 / 0
    )

    with pytest.raises(ZeroDivisionError):
        genconverter.structure("0", PositiveIntNewType)

    assert genconverter.structure("1", PositiveIntNewType) == 1

    with pytest.raises(ZeroDivisionError):
        genconverter.structure("0", BigPositiveIntNewType)

    genconverter.register_structure_hook(
        BigPositiveIntNewType, lambda v, _: int(v) if int(v) > 50 else 1 / 0
    )

    with pytest.raises(ZeroDivisionError):
        genconverter.structure("1", BigPositiveIntNewType)

    assert genconverter.structure("1", PositiveIntNewType) == 1
    assert genconverter.structure("51", BigPositiveIntNewType) == 51


def test_newtype_unstructure_hooks(genconverter: BaseConverter):
    """NewTypes should work with `register_unstructure_hook`."""

    assert genconverter.unstructure(0, int) == 0
    assert genconverter.unstructure(0, PositiveIntNewType) == 0
    assert genconverter.unstructure(0, BigPositiveIntNewType) == 0

    genconverter.register_unstructure_hook(PositiveIntNewType, oct)

    assert genconverter.unstructure(0, PositiveIntNewType) == "0o0"
    assert genconverter.unstructure(0, BigPositiveIntNewType) == "0o0"

    genconverter.register_unstructure_hook(BigPositiveIntNewType, hex)

    assert genconverter.unstructure(0, PositiveIntNewType) == "0o0"
    assert genconverter.unstructure(0, BigPositiveIntNewType) == "0x0"



from collections import Counter
from collections.abc import (
    Mapping,
    MutableMapping,
    MutableSequence,
    MutableSet,
    Sequence,
    Set,
)
from functools import partial

import attr
import pytest
from immutables import Map

from cattrs import Converter
from cattrs._compat import is_py39_plus
from cattrs.converters import is_mutable_set, is_sequence


@pytest.mark.skipif(not is_py39_plus, reason="Requires Python 3.9+")
def test_collection_unstructure_override_set():
    """Test overriding unstructuring sets."""

    # First approach, predicate hook with is_mutable_set
    c = Converter()

    c._unstructure_func.register_func_list(
        [
            (
                is_mutable_set,
                partial(c.gen_unstructure_iterable, unstructure_to=list),
                True,
            )
        ]
    )

    assert c.unstructure({1, 2, 3}, unstructure_as=Set[int]) == [1, 2, 3]

    # Second approach, using __builtins__.set
    c = Converter(unstruct_collection_overrides={set: list})

    assert c.unstructure({1, 2, 3}, unstructure_as=Set[int]) == {1, 2, 3}
    assert c.unstructure({1, 2, 3}, unstructure_as=MutableSet[int]) == {1, 2, 3}
    assert c.unstructure({1, 2, 3}) == [1, 2, 3]

    # Second approach, using abc.MutableSet
    c = Converter(unstruct_collection_overrides={MutableSet: list})

    assert c.unstructure({1, 2, 3}, unstructure_as=Set[int]) == {1, 2, 3}
    assert c.unstructure({1, 2, 3}, unstructure_as=MutableSet[int]) == [1, 2, 3]
    assert c.unstructure({1, 2, 3}) == [1, 2, 3]

    # Second approach, using abc.Set
    c = Converter(unstruct_collection_overrides={Set: list})

    assert c.unstructure({1, 2, 3}, unstructure_as=Set[int]) == [1, 2, 3]
    assert c.unstructure({1, 2, 3}, unstructure_as=MutableSet[int]) == [1, 2, 3]
    assert c.unstructure({1, 2, 3}) == [1, 2, 3]


@pytest.mark.skipif(is_py39_plus, reason="Requires Python 3.8 or lower")
def test_collection_unstructure_override_set_38():
    """Test overriding unstructuring sets."""
    from typing import AbstractSet, MutableSet, Set

    # First approach, predicate hook with is_mutable_set
    c = Converter()

    c._unstructure_func.register_func_list(
        [
            (
                is_mutable_set,
                partial(c.gen_unstructure_iterable, unstructure_to=list),
                True,
            )
        ]
    )

    assert c.unstructure({1, 2, 3}, unstructure_as=Set[int]) == [1, 2, 3]

    # Second approach, using __builtins__.set
    c = Converter(unstruct_collection_overrides={set: list})

    assert c.unstructure({1, 2, 3}, unstructure_as=Set[int]) == [1, 2, 3]
    assert c.unstructure({1, 2, 3}, unstructure_as=MutableSet[int]) == {1, 2, 3}
    assert c.unstructure({1, 2, 3}) == [1, 2, 3]

    # Second approach, using typing.MutableSet
    c = Converter(unstruct_collection_overrides={MutableSet: list})

    assert c.unstructure({1, 2, 3}, unstructure_as=Set[int]) == [1, 2, 3]
    assert c.unstructure({1, 2, 3}, unstructure_as=MutableSet[int]) == [1, 2, 3]
    assert c.unstructure({1, 2, 3}) == [1, 2, 3]

    # Second approach, using typing.AbstractSet
    c = Converter(unstruct_collection_overrides={AbstractSet: list})

    assert c.unstructure({1, 2, 3}, unstructure_as=Set[int]) == [1, 2, 3]
    assert c.unstructure({1, 2, 3}, unstructure_as=MutableSet[int]) == [1, 2, 3]
    assert c.unstructure({1, 2, 3}) == [1, 2, 3]


@pytest.mark.skipif(not is_py39_plus, reason="Requires Python 3.9+")
def test_collection_unstructure_override_seq():
    """Test overriding unstructuring seq."""

    # First approach, predicate hook
    c = Converter()

    c._unstructure_func.register_func_list(
        [(is_sequence, partial(c.gen_unstructure_iterable, unstructure_to=tuple), True)]
    )

    assert c.unstructure([1, 2, 3], unstructure_as=Sequence[int]) == (1, 2, 3)

    @attr.define
    class MyList:
        args = attr.ib(converter=list)

    # Second approach, using abc.MutableSequence
    c = Converter(unstruct_collection_overrides={MutableSequence: MyList})

    assert c.unstructure([1, 2, 3], unstructure_as=Sequence[int]) == [1, 2, 3]
    assert c.unstructure([1, 2, 3], unstructure_as=MutableSequence[int]) == MyList(
        [1, 2, 3]
    )
    assert c.unstructure([1, 2, 3]) == MyList([1, 2, 3])
    assert c.unstructure((1, 2, 3)) == [1, 2, 3]

    # Second approach, using abc.Sequence
    c = Converter(unstruct_collection_overrides={Sequence: MyList})

    assert c.unstructure([1, 2, 3], unstructure_as=Sequence[int]) == MyList([1, 2, 3])
    assert c.unstructure([1, 2, 3], unstructure_as=MutableSequence[int]) == MyList(
        [1, 2, 3]
    )

    assert c.unstructure([1, 2, 3]) == MyList([1, 2, 3])

    assert c.unstructure((1, 2, 3), unstructure_as=tuple[int, ...]) == MyList([1, 2, 3])

    # Second approach, using __builtins__.list
    c = Converter(unstruct_collection_overrides={list: MyList})

    assert c.unstructure([1, 2, 3], unstructure_as=Sequence[int]) == [1, 2, 3]
    assert c.unstructure([1, 2, 3], unstructure_as=MutableSequence[int]) == [1, 2, 3]
    assert c.unstructure([1, 2, 3]) == MyList([1, 2, 3])
    assert c.unstructure((1, 2, 3)) == [1, 2, 3]

    # Second approach, using __builtins__.tuple
    c = Converter(unstruct_collection_overrides={tuple: MyList})

    assert c.unstructure([1, 2, 3], unstructure_as=Sequence[int]) == [1, 2, 3]
    assert c.unstructure([1, 2, 3], unstructure_as=MutableSequence[int]) == [1, 2, 3]
    assert c.unstructure([1, 2, 3]) == [1, 2, 3]
    assert c.unstructure((1, 2, 3)) == MyList([1, 2, 3])


@pytest.mark.skipif(not is_py39_plus, reason="Requires Python 3.9+")
def test_collection_unstructure_override_mapping():
    """Test overriding unstructuring mappings."""

    # Using Counter
    c = Converter(unstruct_collection_overrides={Counter: Map})
    assert c.unstructure(Counter({1: 2})) == Map({1: 2})
    assert c.unstructure(Counter({1: 2}), unstructure_as=Counter[int]) == Map({1: 2})
    assert c.unstructure({1: 2}) == {1: 2}
    assert c.unstructure({1: 2}, unstructure_as=MutableMapping[int, int]) == {1: 2}
    assert c.unstructure({1: 2}, unstructure_as=Mapping[int, int]) == {1: 2}

    # Using __builtins__.dict
    c = Converter(unstruct_collection_overrides={dict: Map})

    assert c.unstructure(Counter({1: 2})) == Map({1: 2})
    assert c.unstructure(Counter({1: 2}), unstructure_as=Counter[int]) == Map({1: 2})
    assert c.unstructure({1: 2}) == Map({1: 2})
    assert c.unstructure({1: 2}, unstructure_as=MutableMapping[int, int]) == {1: 2}
    assert c.unstructure({1: 2}, unstructure_as=Mapping[int, int]) == {1: 2}

    # Using MutableMapping
    c = Converter(unstruct_collection_overrides={MutableMapping: Map})

    assert c.unstructure(Counter({1: 2})) == Map({1: 2})
    assert c.unstructure(Counter({1: 2}), unstructure_as=Counter[int]) == Map({1: 2})
    assert c.unstructure({1: 2}) == Map({1: 2})
    assert c.unstructure({1: 2}, unstructure_as=MutableMapping[int, int]) == Map({1: 2})
    assert c.unstructure({1: 2}, unstructure_as=Mapping[int, int]) == {1: 2}

    # Using Mapping
    c = Converter(unstruct_collection_overrides={Mapping: Map})

    assert c.unstructure(Counter({1: 2})) == Map({1: 2})
    assert c.unstructure(Counter({1: 2}), unstructure_as=Counter[int]) == Map({1: 2})
    assert c.unstructure({1: 2}) == Map({1: 2})
    assert c.unstructure({1: 2}, unstructure_as=MutableMapping[int, int]) == Map({1: 2})
    assert c.unstructure({1: 2}, unstructure_as=Mapping[int, int]) == Map({1: 2})
